"""Tests for the loggers module."""

from unittest.mock import MagicMock

import pytest

from mkdocstrings import get_logger, get_template_logger


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"once": False},
        {"once": True},
    ],
)
def test_logger(kwargs: dict, caplog: pytest.LogCaptureFixture) -> None:
    """Test logger methods.

    Parameters:
        kwargs: Keyword arguments passed to the logger methods.
    """
    logger = get_logger("mkdocstrings.test")
    caplog.set_level(0)
    for _ in range(2):
        logger.debug("Debug message", **kwargs)
        logger.info("Info message", **kwargs)
        logger.warning("Warning message", **kwargs)
        logger.error("Error message", **kwargs)
        logger.critical("Critical message", **kwargs)
    if kwargs.get("once", False):
        assert len(caplog.records) == 5
    else:
        assert len(caplog.records) == 10


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"once": False},
        {"once": True},
    ],
)
def test_template_logger(kwargs: dict, caplog: pytest.LogCaptureFixture) -> None:
    """Test template logger methods.

    Parameters:
        kwargs: Keyword arguments passed to the template logger methods.
    """
    logger = get_template_logger()
    mock = MagicMock()
    caplog.set_level(0)
    for _ in range(2):
        logger.debug(mock, "Debug message", **kwargs)
        logger.info(mock, "Info message", **kwargs)
        logger.warning(mock, "Warning message", **kwargs)
        logger.error(mock, "Error message", **kwargs)
        logger.critical(mock, "Critical message", **kwargs)
    if kwargs.get("once", False):
        assert len(caplog.records) == 5
    else:
        assert len(caplog.records) == 10



"""Configuration for the pytest test suite."""

from __future__ import annotations

from collections import ChainMap
from typing import TYPE_CHECKING, Any

import pytest
from markdown.core import Markdown
from mkdocs.config.defaults import MkDocsConfig

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from mkdocs import config

    from mkdocstrings._internal.plugin import MkdocstringsPlugin


@pytest.fixture(name="mkdocs_conf")
def fixture_mkdocs_conf(request: pytest.FixtureRequest, tmp_path: Path) -> Iterator[config.Config]:
    """Yield a MkDocs configuration object."""
    conf = MkDocsConfig()
    while hasattr(request, "_parent_request") and hasattr(request._parent_request, "_parent_request"):
        request = request._parent_request  # ty: ignore[invalid-assignment]

    conf_dict = {
        "site_name": "foo",
        "site_url": "https://example.org/",
        "site_dir": str(tmp_path),
        "plugins": [{"mkdocstrings": {"default_handler": "python"}}],
        **getattr(request, "param", {}),
    }
    # Re-create it manually as a workaround for https://github.com/mkdocs/mkdocs/issues/2289
    mdx_configs: dict[str, Any] = dict(ChainMap(*conf_dict.get("markdown_extensions", [])))  # ty: ignore[invalid-argument-type,invalid-assignment]

    conf.load_dict(conf_dict)
    assert conf.validate() == ([], [])

    conf["mdx_configs"] = mdx_configs
    conf["markdown_extensions"].insert(0, "toc")  # Guaranteed to be added by MkDocs.

    conf = conf["plugins"]["mkdocstrings"].on_config(conf)
    conf = conf["plugins"]["autorefs"].on_config(conf)
    yield conf
    conf["plugins"]["mkdocstrings"].on_post_build(conf)


@pytest.fixture(name="plugin")
def fixture_plugin(mkdocs_conf: config.Config) -> MkdocstringsPlugin:
    """Return a plugin instance."""
    return mkdocs_conf["plugins"]["mkdocstrings"]


@pytest.fixture(name="ext_markdown")
def fixture_ext_markdown(mkdocs_conf: MkDocsConfig) -> Markdown:
    """Return a Markdown instance with MkdocstringsExtension."""
    return Markdown(extensions=mkdocs_conf["markdown_extensions"], extension_configs=mkdocs_conf["mdx_configs"])



"""Tests for the handlers.base module."""

from __future__ import annotations

from textwrap import dedent
from typing import TYPE_CHECKING

import pytest
from dirty_equals import IsStr
from jinja2.exceptions import TemplateNotFound
from markdown import Markdown

from mkdocstrings import Highlighter

if TYPE_CHECKING:
    from pathlib import Path

    from mkdocstrings import MkdocstringsPlugin


@pytest.mark.parametrize("extension_name", ["codehilite", "pymdownx.highlight"])
def test_highlighter_without_pygments(extension_name: str) -> None:
    """Assert that it's possible to disable Pygments highlighting.

    Arguments:
        extension_name: The "user-chosen" Markdown extension for syntax highlighting.
    """
    configs = {extension_name: {"use_pygments": False, "css_class": "hiiii"}}
    md = Markdown(extensions=[extension_name], extension_configs=configs)
    hl = Highlighter(md)
    assert (
        hl.highlight("import foo", language="python")
        == '<pre class="hiiii"><code class="language-python">import foo</code></pre>'
    )
    assert (
        hl.highlight("import foo", language="python", inline=True)
        == f'<code class="{"" if extension_name == "codehilite" else "hiiii"} language-python">import foo</code>'
    )


@pytest.mark.parametrize("extension_name", [None, "codehilite", "pymdownx.highlight"])
@pytest.mark.parametrize("inline", [False, True])
def test_highlighter_basic(extension_name: str | None, inline: bool) -> None:
    """Assert that Pygments syntax highlighting works.

    Arguments:
        extension_name: The "user-chosen" Markdown extension for syntax highlighting.
        inline: Whether the highlighting was inline.
    """
    md = Markdown(extensions=[extension_name], extension_configs={extension_name: {}}) if extension_name else Markdown()
    hl = Highlighter(md)

    actual = hl.highlight("import foo", language="python", inline=inline)
    assert "import" in actual
    assert "import foo" not in actual  # Highlighting has split it up.


def test_extended_templates(tmp_path: Path, plugin: MkdocstringsPlugin) -> None:
    """Test the extended templates functionality.

    Parameters:
        tmp_path: Temporary folder.
        plugin: Instance of our plugin.
    """
    handler = plugin._handlers.get_handler("python")  # ty: ignore[unresolved-attribute]

    # monkeypatch Jinja env search path
    search_paths = [
        base_theme := tmp_path / "base_theme",
        base_fallback_theme := tmp_path / "base_fallback_theme",
        extended_theme := tmp_path / "extended_theme",
        extended_fallback_theme := tmp_path / "extended_fallback_theme",
    ]
    handler.env.loader.searchpath = search_paths  # ty: ignore[invalid-assignment]

    # assert "new" template is not found
    with pytest.raises(expected_exception=TemplateNotFound):
        handler.env.get_template("new.html")

    # check precedence: base theme, base fallback theme, extended theme, extended fallback theme
    # start with last one and go back up
    handler.env.cache = None

    extended_fallback_theme.mkdir()
    extended_fallback_theme.joinpath("new.html").write_text("extended fallback new")
    assert handler.env.get_template("new.html").render() == "extended fallback new"

    extended_theme.mkdir()
    extended_theme.joinpath("new.html").write_text("extended new")
    assert handler.env.get_template("new.html").render() == "extended new"

    base_fallback_theme.mkdir()
    base_fallback_theme.joinpath("new.html").write_text("base fallback new")
    assert handler.env.get_template("new.html").render() == "base fallback new"

    base_theme.mkdir()
    base_theme.joinpath("new.html").write_text("base new")
    assert handler.env.get_template("new.html").render() == "base new"


@pytest.mark.parametrize(
    "ext_markdown",
    [{"markdown_extensions": [{"toc": {"permalink": True}}]}],
    indirect=["ext_markdown"],
)
def test_nested_autodoc(ext_markdown: Markdown) -> None:
    """Assert that nested autodocs render well and do not mess up the TOC."""
    output = ext_markdown.convert(
        dedent(
            """
            # ::: tests.fixtures.nesting.Class
                options:
                    members: false
                    show_root_heading: true
            """,
        ),
    )
    assert 'id="tests.fixtures.nesting.Class"' in output
    assert 'id="tests.fixtures.nesting.Class.method"' in output
    assert ext_markdown.toc_tokens == [  # ty: ignore[unresolved-attribute]
        {
            "level": 1,
            "id": "tests.fixtures.nesting.Class",
            "html": IsStr(),
            "name": "Class",
            "data-toc-label": "Class",
            "children": [
                {
                    "level": 2,
                    "id": "tests.fixtures.nesting.Class.method",
                    "html": IsStr(),
                    "name": "method",
                    "data-toc-label": "method",
                    "children": [],
                },
            ],
        },
    ]



# Logging functions.

from __future__ import annotations

import logging
from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING, Any

from jinja2 import pass_context

if TYPE_CHECKING:
    from collections.abc import Callable, MutableMapping, Sequence

    from jinja2.runtime import Context


try:
    import mkdocstrings_handlers
except ImportError:
    TEMPLATES_DIRS: Sequence[Path] = ()
else:
    TEMPLATES_DIRS = tuple(mkdocstrings_handlers.__path__)
    """The directories where the handler templates are located."""


class LoggerAdapter(logging.LoggerAdapter):
    """A logger adapter to prefix messages.

    This adapter also adds an additional parameter to logging methods
    called `once`: if `True`, the message will only be logged once.

    Examples:
        In Python code:

        >>> logger = get_logger("myplugin")
        >>> logger.debug("This is a debug message.")
        >>> logger.info("This is an info message.", once=True)

        In Jinja templates (logger available in context as `log`):

        ```jinja
        {{ log.debug("This is a debug message.") }}
        {{ log.info("This is an info message.", once=True) }}
        ```
    """

    def __init__(self, prefix: str, logger: logging.Logger):
        """Initialize the object.

        Arguments:
            prefix: The string to insert in front of every message.
            logger: The logger instance.
        """
        super().__init__(logger, {})
        self.prefix = prefix
        """The prefix to insert in front of every message."""
        self._logged: set[tuple[LoggerAdapter, str]] = set()

    def process(self, msg: str, kwargs: MutableMapping[str, Any]) -> tuple[str, Any]:
        """Process the message.

        Arguments:
            msg: The message:
            kwargs: Remaining arguments.

        Returns:
            The processed message.
        """
        return f"{self.prefix}: {msg}", kwargs

    def log(self, level: int, msg: object, *args: object, **kwargs: object) -> None:
        """Log a message.

        Arguments:
            level: The logging level.
            msg: The message.
            *args: Additional arguments passed to parent method.
            **kwargs: Additional keyword arguments passed to parent method.
        """
        if kwargs.pop("once", False):
            if (key := (self, str(msg))) in self._logged:
                return
            self._logged.add(key)
        super().log(level, msg, *args, **kwargs)  # ty: ignore[invalid-argument-type]


class TemplateLogger:
    """A wrapper class to allow logging in templates.

    The logging methods provided by this class all accept
    two parameters:

    - `msg`: The message to log.
    - `once`: If `True`, the message will only be logged once.

    Methods:
        debug: Function to log a DEBUG message.
        info: Function to log an INFO message.
        warning: Function to log a WARNING message.
        error: Function to log an ERROR message.
        critical: Function to log a CRITICAL message.
    """

    def __init__(self, logger: LoggerAdapter):
        """Initialize the object.

        Arguments:
            logger: A logger adapter.
        """
        self.debug = get_template_logger_function(logger.debug)
        """Log a DEBUG message."""
        self.info = get_template_logger_function(logger.info)
        """Log an INFO message."""
        self.warning = get_template_logger_function(logger.warning)
        """Log a WARNING message."""
        self.error = get_template_logger_function(logger.error)
        """Log an ERROR message."""
        self.critical = get_template_logger_function(logger.critical)
        """Log a CRITICAL message."""


class _Lazy:
    unset = object()

    def __init__(self, func: Callable, *args: Any, **kwargs: Any):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.result = self.unset

    def __call__(self):
        if self.result is self.unset:
            self.result = self.func(*self.args, **self.kwargs)
        return self.result

    def __str__(self) -> str:
        return str(self())

    def __repr__(self) -> str:
        return repr(self())


def get_template_logger_function(logger_func: Callable) -> Callable:
    """Create a wrapper function that automatically receives the Jinja template context.

    Arguments:
        logger_func: The logger function to use within the wrapper.

    Returns:
        A function.
    """

    @pass_context
    def wrapper(context: Context, msg: str | None = None, *args: Any, **kwargs: Any) -> str:
        """Log a message.

        Arguments:
            context: The template context, automatically provided by Jinja.
            msg: The message to log.
            **kwargs: Additional arguments passed to the logger function.

        Returns:
            An empty string.
        """
        logger_func(f"%s: {msg or 'Rendering'}", _Lazy(get_template_path, context), *args, **kwargs)
        return ""

    return wrapper


def get_template_path(context: Context) -> str:
    """Return the path to the template currently using the given context.

    Arguments:
        context: The template context.

    Returns:
        The relative path to the template.
    """
    context_name: str = str(context.name)
    filename = context.environment.get_template(context_name).filename
    if filename:
        for template_dir in TEMPLATES_DIRS:
            with suppress(ValueError):
                return str(Path(filename).relative_to(template_dir))
        with suppress(ValueError):
            return str(Path(filename).relative_to(Path.cwd()))
        return filename
    return context_name


def get_logger(name: str) -> LoggerAdapter:
    """Return a pre-configured logger.

    Arguments:
        name: The name to use with `logging.getLogger`.

    Returns:
        A logger configured to work well in MkDocs.
    """
    logger = logging.getLogger(f"mkdocs.plugins.{name}")
    return LoggerAdapter(name.split(".", 1)[0], logger)


def get_template_logger(handler_name: str | None = None) -> TemplateLogger:
    """Return a logger usable in templates.

    Parameters:
        handler_name: The name of the handler.

    Returns:
        A template logger.
    """
    handler_name = handler_name or "base"
    return TemplateLogger(get_logger(f"mkdocstrings_handlers.{handler_name}.templates"))



# Base module for handlers.
#
# This module contains the base classes for implementing handlers.

from __future__ import annotations

import datetime
import importlib
import ssl
from concurrent import futures
from importlib.metadata import entry_points
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO, ClassVar, cast
from warnings import warn
from xml.etree.ElementTree import Element, tostring

from jinja2 import Environment, FileSystemLoader
from markdown import Markdown
from markupsafe import Markup
from mkdocs.utils.cache import download_and_cache_url
from mkdocs_autorefs import AutorefsInlineProcessor, BacklinksTreeProcessor

from mkdocstrings._internal.download import _download_url_with_gz
from mkdocstrings._internal.handlers.rendering import (
    HeadingShiftingTreeprocessor,
    Highlighter,
    IdPrependingTreeprocessor,
    MkdocstringsInnerExtension,
    ParagraphStrippingTreeprocessor,
)
from mkdocstrings._internal.inventory import Inventory
from mkdocstrings._internal.loggers import get_logger, get_template_logger

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping, Sequence

    from markdown import Extension
    from markdown.extensions.toc import TocTreeprocessor
    from mkdocs_autorefs import AutorefsHookInterface, Backlink

_logger = get_logger("mkdocstrings")

CollectorItem = Any
"""The type of the item returned by the `collect` method of a handler."""
HandlerConfig = Any
"""The type of the configuration of a handler."""
HandlerOptions = Any
"""The type of the options passed to a handler."""


# Autodoc instructions can appear in nested Markdown,
# so we need to keep track of the Markdown conversion layer we're in.
# Since any handler can be called from any Markdown conversion layer,
# we need to keep track of the layer globally.
# This global variable is incremented/decremented in `do_convert_markdown`,
# and used in `outer_layer`.
_markdown_conversion_layer: int = 0


class CollectionError(Exception):
    """An exception raised when some collection of data failed."""


class ThemeNotSupported(Exception):  # noqa: N818
    """An exception raised to tell a theme is not supported."""


def do_any(seq: Sequence, attribute: str | None = None) -> bool:
    """Check if at least one of the item in the sequence evaluates to true.

    The `any` builtin as a filter for Jinja templates.

    Arguments:
        seq: An iterable object.
        attribute: The attribute name to use on each object of the iterable.

    Returns:
        A boolean telling if any object of the iterable evaluated to True.
    """
    if attribute is None:
        return any(seq)
    return any(_[attribute] for _ in seq)


class BaseHandler:
    """The base handler class.

    Inherit from this class to implement a handler.

    You will have to implement the `collect` and `render` methods.
    You can also implement the `teardown` method,
    and  override the `update_env` method, to add more filters to the Jinja environment,
    making them available in your Jinja templates.

    To define a fallback theme, add a `fallback_theme` class-variable.
    To add custom CSS, add an `extra_css` variable or create an 'style.css' file beside the templates.
    """

    name: ClassVar[str]
    """The handler's name, for example "python"."""

    domain: ClassVar[str]
    """The handler's domain, used to register objects in the inventory, for example "py"."""

    enable_inventory: ClassVar[bool] = False
    """Whether the inventory creation is enabled."""

    fallback_theme: ClassVar[str] = ""
    """Fallback theme to use when a template isn't found in the configured theme."""

    extra_css: str = ""
    """Extra CSS."""

    def __init__(
        self,
        *,
        theme: str,
        custom_templates: str | None,
        mdx: Sequence[str | Extension],
        mdx_config: Mapping[str, Any],
    ) -> None:
        """Initialize the object.

        If the given theme is not supported (it does not exist), it will look for a `fallback_theme` attribute
        in `self` to use as a fallback theme.

        Keyword Arguments:
            theme (str): The theme to use.
            custom_templates (str | None): The path to custom templates.
            mdx (list[str | Extension]): A list of Markdown extensions to use.
            mdx_config (Mapping[str, Mapping[str, Any]]): Configuration for the Markdown extensions.
        """
        self.theme = theme
        """The selected theme."""
        self.custom_templates = custom_templates
        """The path to custom templates."""
        self.mdx = mdx
        """The Markdown extensions to use."""
        self.mdx_config = mdx_config
        """The configuration for the Markdown extensions."""
        self._md: Markdown | None = None
        self._headings: list[Element] = []

        paths = []

        # add selected theme templates
        themes_dir = self.get_templates_dir(self.name)
        paths.append(themes_dir / self.theme)

        # add extended theme templates
        extended_templates_dirs = self.get_extended_templates_dirs(self.name)
        paths.extend(templates_dir / self.theme for templates_dir in extended_templates_dirs)

        # add fallback theme templates
        if self.fallback_theme and self.fallback_theme != self.theme:
            paths.append(themes_dir / self.fallback_theme)

            # add fallback theme of extended templates
            paths.extend(templates_dir / self.fallback_theme for templates_dir in extended_templates_dirs)

        for path in paths:
            css_path = path / "style.css"
            if css_path.is_file():
                self.extra_css += "\n" + css_path.read_text(encoding="utf-8")
                break

        if self.custom_templates is not None:
            paths.insert(0, Path(self.custom_templates) / self.name / self.theme)

        self.env = Environment(
            autoescape=True,
            loader=FileSystemLoader(paths),
            auto_reload=False,  # Editing a template in the middle of a build is not useful.
        )
        """The Jinja environment."""

        self.env.filters["convert_markdown"] = self.do_convert_markdown
        self.env.filters["heading"] = self.do_heading
        self.env.filters["any"] = do_any
        self.env.globals["log"] = get_template_logger(self.name)  # ty:ignore[invalid-assignment]

    @property
    def md(self) -> Markdown:
        """The Markdown instance.

        Raises:
            RuntimeError: When the Markdown instance is not set yet.
        """
        if self._md is None:
            raise RuntimeError("Markdown instance not set yet")
        return self._md

    def get_inventory_urls(self) -> list[tuple[str, dict[str, Any]]]:
        """Return the URLs (and configuration options) of the inventory files to download."""
        return []

    @classmethod
    def load_inventory(
        cls,
        in_file: BinaryIO,  # noqa: ARG003
        url: str,  # noqa: ARG003
        base_url: str | None = None,  # noqa: ARG003
        **kwargs: Any,  # noqa: ARG003
    ) -> Iterator[tuple[str, str]]:
        """Yield items and their URLs from an inventory file streamed from `in_file`.

        Arguments:
            in_file: The binary file-like object to read the inventory from.
            url: The URL that this file is being streamed from (used to guess `base_url`).
            base_url: The URL that this inventory's sub-paths are relative to.
            **kwargs: Ignore additional arguments passed from the config.

        Yields:
            Tuples of (item identifier, item URL).
        """
        yield from ()

    def get_options(self, local_options: Mapping[str, Any]) -> HandlerOptions:
        """Get combined options.

        Override this method to customize how options are combined,
        for example by merging the global options with the local options.
        By combining options here, you don't have to do it twice in `collect` and `render`.

        Arguments:
            local_options: The local options.

        Returns:
            The combined options.
        """
        return local_options

    def collect(self, identifier: str, options: HandlerOptions) -> CollectorItem:
        """Collect data given an identifier and user configuration.

        In the implementation, you typically call a subprocess that returns JSON, and load that JSON again into
        a Python dictionary for example, though the implementation is completely free.

        Arguments:
            identifier: An identifier for which to collect data. For example, in Python,
                it would be 'mkdocstrings.BaseHandler' to collect documentation about the BaseHandler class.
                It can be anything that you can feed to the tool of your choice.
            options: The final configuration options.

        Returns:
            Anything you want, as long as you can feed it to the handler's `render` method.
        """
        raise NotImplementedError

    def render(self, data: CollectorItem, options: HandlerOptions, *, locale: str | None = None) -> str:
        """Render a template using provided data and configuration options.

        Arguments:
            data: The collected data to render.
            options: The final configuration options.
            locale: The locale to use for translations, if any.

        Returns:
            The rendered template as HTML.
        """
        raise NotImplementedError

    def render_backlinks(self, backlinks: Mapping[str, Iterable[Backlink]], *, locale: str | None = None) -> str:  # noqa: ARG002
        """Render backlinks.

        Parameters:
            backlinks: A mapping of identifiers to backlinks.
            locale: The locale to use for translations, if any.

        Returns:
            The rendered backlinks as HTML.
        """
        return ""

    def teardown(self) -> None:
        """Teardown the handler.

        This method should be implemented to, for example, terminate a subprocess
        that was started when creating the handler instance.
        """

    def get_templates_dir(self, handler: str | None = None) -> Path:
        """Return the path to the handler's templates directory.

        Override to customize how the templates directory is found.

        Arguments:
            handler: The name of the handler to get the templates directory of.

        Raises:
            ModuleNotFoundError: When no such handler is installed.
            FileNotFoundError: When the templates directory cannot be found.

        Returns:
            The templates directory path.
        """
        handler = handler or self.name
        try:
            import mkdocstrings_handlers  # noqa: PLC0415
        except ModuleNotFoundError as error:
            raise ModuleNotFoundError(f"Handler '{handler}' not found, is it installed?") from error

        for path in mkdocstrings_handlers.__path__:
            theme_path = Path(path, handler, "templates")
            if theme_path.exists():
                return theme_path

        raise FileNotFoundError(f"Can't find 'templates' folder for handler '{handler}'")

    def get_extended_templates_dirs(self, handler: str) -> list[Path]:
        """Load template extensions for the given handler, return their templates directories.

        Arguments:
            handler: The name of the handler to get the extended templates directory of.

        Returns:
            The extensions templates directories.
        """
        discovered_extensions = entry_points(group=f"mkdocstrings.{handler}.templates")
        return [extension.load()() for extension in discovered_extensions]

    def get_aliases(self, identifier: str) -> tuple[str, ...]:  # noqa: ARG002
        """Return the possible aliases for a given identifier.

        Arguments:
            identifier: The identifier to get the aliases of.

        Returns:
            A tuple of strings - aliases.
        """
        return ()

    @property
    def outer_layer(self) -> bool:
        """Whether we're in the outer Markdown conversion layer."""
        return _markdown_conversion_layer == 0

    def do_convert_markdown(
        self,
        text: str,
        heading_level: int,
        html_id: str = "",
        *,
        strip_paragraph: bool = False,
        autoref_hook: AutorefsHookInterface | None = None,
    ) -> Markup:
        """Render Markdown text; for use inside templates.

        Arguments:
            text: The text to convert.
            heading_level: The base heading level to start all Markdown headings from.
            html_id: The HTML id of the element that's considered the parent of this element.
            strip_paragraph: Whether to exclude the `<p>` tag from around the whole output.

        Returns:
            An HTML string.
        """
        global _markdown_conversion_layer  # noqa: PLW0603
        _markdown_conversion_layer += 1
        treeprocessors = self.md.treeprocessors
        treeprocessors[HeadingShiftingTreeprocessor.name].shift_by = heading_level
        treeprocessors[IdPrependingTreeprocessor.name].id_prefix = html_id and html_id + "--"
        treeprocessors[ParagraphStrippingTreeprocessor.name].strip = strip_paragraph
        if BacklinksTreeProcessor.name in treeprocessors:
            treeprocessors[BacklinksTreeProcessor.name].initial_id = html_id
        if autoref_hook and AutorefsInlineProcessor.name in self.md.inlinePatterns:
            self.md.inlinePatterns[AutorefsInlineProcessor.name].hook = autoref_hook  # ty: ignore[unresolved-attribute]

        try:
            return Markup(self.md.convert(text))
        finally:
            treeprocessors[HeadingShiftingTreeprocessor.name].shift_by = 0
            treeprocessors[IdPrependingTreeprocessor.name].id_prefix = ""
            treeprocessors[ParagraphStrippingTreeprocessor.name].strip = False
            if BacklinksTreeProcessor.name in treeprocessors:
                treeprocessors[BacklinksTreeProcessor.name].initial_id = None
            if AutorefsInlineProcessor.name in self.md.inlinePatterns:
                self.md.inlinePatterns[AutorefsInlineProcessor.name].hook = None  # ty: ignore[unresolved-attribute]
            self.md.reset()
            _markdown_conversion_layer -= 1

    def do_heading(
        self,
        content: Markup,
        heading_level: int,
        *,
        role: str | None = None,
        hidden: bool = False,
        toc_label: str | None = None,
        skip_inventory: bool = False,
        **attributes: str,
    ) -> Markup:
        """Render an HTML heading and register it for the table of contents. For use inside templates.

        Arguments:
            content: The HTML within the heading.
            heading_level: The level of heading (e.g. 3 -> `h3`).
            role: An optional role for the object bound to this heading.
            hidden: If True, only register it for the table of contents, don't render anything.
            toc_label: The title to use in the table of contents ('data-toc-label' attribute).
            skip_inventory: Flag element to not be registered in the inventory (by setting a `data-skip-inventory` attribute).
            **attributes: Any extra HTML attributes of the heading.

        Returns:
            An HTML string.
        """
        # Produce a heading element that will be used later, in `AutoDocProcessor.run`, to:
        # - register it in the ToC: right now we're in the inner Markdown conversion layer,
        #   so we have to bubble up the information to the outer Markdown conversion layer,
        #   for the ToC extension to pick it up.
        # - register it in autorefs: right now we don't know what page is being rendered,
        #   so we bubble up the information again to where autorefs knows the page,
        #   and can correctly register the heading anchor (id) to its full URL.
        # - register it in the objects inventory: same as for autorefs,
        #   we don't know the page here, or the handler (and its domain),
        #   so we bubble up the information to where the mkdocstrings extension knows that.
        el = Element(f"h{heading_level}", attributes)
        if toc_label is None:
            toc_label = content.unescape() if isinstance(content, Markup) else content
        el.set("data-toc-label", toc_label)
        if skip_inventory:
            el.set("data-skip-inventory", "true")
        if role:
            el.set("data-role", role)
        if content:
            el.text = str(content).strip()
        self._headings.append(el)

        if hidden:
            return Markup('<a id="{0}"></a>').format(attributes["id"])

        # Now produce the actual HTML to be rendered. The goal is to wrap the HTML content into a heading.
        # Start with a heading that has just attributes (no text), and add a placeholder into it.
        el = Element(f"h{heading_level}", attributes)
        el.append(Element("mkdocstrings-placeholder"))
        # Tell the inner 'toc' extension to make its additions if configured so.
        toc = cast("TocTreeprocessor", self.md.treeprocessors["toc"])
        if toc.use_anchors:
            toc.add_anchor(el, attributes["id"])
        if toc.use_permalinks:
            toc.add_permalink(el, attributes["id"])

        # The content we received is HTML, so it can't just be inserted into the tree. We had marked the middle
        # of the heading with a placeholder that can never occur (text can't directly contain angle brackets).
        # Now this HTML wrapper can be "filled" by replacing the placeholder.
        html_with_placeholder = tostring(el, encoding="unicode")
        assert (  # noqa: S101
            html_with_placeholder.count("<mkdocstrings-placeholder />") == 1
        ), f"Bug in mkdocstrings: failed to replace in {html_with_placeholder!r}"
        html = html_with_placeholder.replace("<mkdocstrings-placeholder />", content)
        return Markup(html)

    def get_headings(self) -> Sequence[Element]:
        """Return and clear the headings gathered so far.

        Returns:
            A list of HTML elements.
        """
        result = list(self._headings)
        self._headings.clear()
        return result

    def update_env(self, config: Any) -> None:
        """Update the Jinja environment."""

    def _update_env(self, md: Markdown, *, config: Any | None = None) -> None:
        """Update our handler to point to our configured Markdown instance, grabbing some of the config from `md`."""
        # Remove extensions that are not useful in the inner conversion layer.
        # Zensical's `preview` is especially problematic because it expects to find
        # the `zrelpath` treeprocessor while this treeprocessor is registered "manually"
        # by Zensical, meaning that mkdocstrings cannot propagate it naturally.
        # Previously we did propagate it ourselves, but Zensical recently started
        # applying its logic onto HTML stashes, meaning the logic was applied twice:
        # once by mkdocstrings-propagated `zrelpath` treeprocessor, and a second time
        # by Zensical's `zrelpath` postprocessor. See zensical/zensical#818.
        remove_names = ("zensical.extensions.preview",)
        mdx = [ext for ext in self.mdx if ext not in remove_names]
        extensions: list[str | Extension] = [*mdx, MkdocstringsInnerExtension(self._headings)]

        new_md = Markdown(extensions=extensions, extension_configs=self.mdx_config)

        # MkDocs adds its own (required) extension that's not part of the config. Propagate it.
        if "relpath" in md.treeprocessors:
            relpath = md.treeprocessors["relpath"]
            new_relpath = type(relpath)(relpath.file, relpath.files, relpath.config)
            new_md.treeprocessors.register(new_relpath, "relpath", priority=0)

        self._md = new_md

        self.env.filters["highlight"] = Highlighter(new_md).highlight

        self.update_env(config)


class Handlers:
    """A collection of handlers.

    Do not instantiate this directly. [The plugin][mkdocstrings.MkdocstringsPlugin] will keep one instance of
    this for the purpose of caching. Use [mkdocstrings.MkdocstringsPlugin.get_handler][] for convenient access.
    """

    def __init__(
        self,
        *,
        theme: str,
        default: str,
        inventory_project: str,
        inventory_version: str = "0.0.0",
        handlers_config: dict[str, HandlerConfig] | None = None,
        custom_templates: str | None = None,
        mdx: Sequence[str | Extension] | None = None,
        mdx_config: Mapping[str, Any] | None = None,
        locale: str = "en",
        tool_config: Any,
    ) -> None:
        """Initialize the object.

        Arguments:
            theme: The theme to use.
            default: The default handler to use.
            inventory_project: The project name to use in the inventory.
            inventory_version: The project version to use in the inventory.
            handlers_config: The handlers configuration.
            custom_templates: The path to custom templates.
            mdx: A list of Markdown extensions to use.
            mdx_config: Configuration for the Markdown extensions.
            locale: The locale to use for translations.
            tool_config: Tool configuration to pass down to handlers.
        """
        self._theme = theme
        self._default = default
        self._handlers_config = handlers_config or {}
        self._custom_templates = custom_templates
        self._mdx = mdx or []
        self._mdx_config = mdx_config or {}
        self._handlers: dict[str, BaseHandler] = {}
        self._locale = locale
        self._tool_config = tool_config

        self.inventory: Inventory = Inventory(project=inventory_project, version=inventory_version)
        """The objects inventory."""

        self._inv_futures: dict[futures.Future, tuple[BaseHandler, str, Any]] = {}

    def get_handler_name(self, config: dict) -> str:
        """Return the handler name defined in an "autodoc" instruction YAML configuration, or the global default handler.

        Arguments:
            config: A configuration dictionary, obtained from YAML below the "autodoc" instruction.

        Returns:
            The name of the handler to use.
        """
        return config.get("handler", self._default)

    def get_handler_config(self, name: str) -> dict:
        """Return the global configuration of the given handler.

        Arguments:
            name: The name of the handler to get the global configuration of.

        Returns:
            The global configuration of the given handler. It can be an empty dictionary.
        """
        return self._handlers_config.get(name, None) or {}

    def get_handler(self, name: str, handler_config: dict | None = None) -> BaseHandler:
        """Get a handler thanks to its name.

        This function dynamically imports a module named "mkdocstrings_handlers.NAME", calls its
        `get_handler` method to get an instance of a handler, and caches it in dictionary.
        It means that during one run (for each reload when serving, or once when building),
        a handler is instantiated only once, and reused for each "autodoc" instruction asking for it.

        Arguments:
            name: The name of the handler. Really, it's the name of the Python module holding it.
            handler_config: Configuration passed to the handler.

        Returns:
            An instance of a subclass of [`BaseHandler`][mkdocstrings.BaseHandler],
                as instantiated by the `get_handler` method of the handler's module.
        """
        if name not in self._handlers:
            if handler_config is None:
                handler_config = self._handlers_config.get(name, {})
            module = importlib.import_module(f"mkdocstrings_handlers.{name}")

            self._handlers[name] = module.get_handler(
                theme=self._theme,
                custom_templates=self._custom_templates,
                mdx=self._mdx,
                mdx_config=self._mdx_config,
                handler_config=handler_config,
                tool_config=self._tool_config,
            )
        return self._handlers[name]

    def _download_inventories(self) -> None:
        """Download an inventory file from an URL.

        Arguments:
            url: The URL of the inventory.
        """
        to_download: list[tuple[BaseHandler, str, Any]] = []

        for handler_name, conf in self._handlers_config.items():
            handler = self.get_handler(handler_name)

            if handler.get_inventory_urls.__func__ is BaseHandler.get_inventory_urls:
                if inv_configs := conf.pop("import", ()):
                    warn(
                        "mkdocstrings v1 will stop handling 'import' in handlers configuration. "
                        "Instead your handler must define a `get_inventory_urls` method "
                        "that returns a list of URLs to download. ",
                        DeprecationWarning,
                        stacklevel=1,
                    )
                    inv_configs = [{"url": inv} if isinstance(inv, str) else inv for inv in inv_configs]
                    inv_configs = [(inv.pop("url"), inv) for inv in inv_configs]
            else:
                inv_configs = handler.get_inventory_urls()

            to_download.extend((handler, url, conf) for url, conf in inv_configs)

        if to_download:
            # YORE: EOL 3.12: Remove block.
            # NOTE: Create context in main thread to fix issue
            # https://github.com/mkdocstrings/mkdocstrings/issues/796.
            _ = ssl.create_default_context()

            thread_pool = futures.ThreadPoolExecutor(4)
            for handler, url, conf in to_download:
                _logger.debug("Downloading inventory from %s", url)
                future = thread_pool.submit(
                    download_and_cache_url,
                    url,
                    datetime.timedelta(days=1),
                    download=_download_url_with_gz,
                )
                self._inv_futures[future] = (handler, url, conf)
            thread_pool.shutdown(wait=False)

    def _yield_inventory_items(self) -> Iterator[tuple[str, str]]:
        if self._inv_futures:
            _logger.debug("Waiting for %s inventory download(s)", len(self._inv_futures))
            futures.wait(self._inv_futures, timeout=30)
            # Reversed order so that pages from first futures take precedence:
            for fut, (handler, url, conf) in reversed(self._inv_futures.items()):
                try:
                    yield from handler.load_inventory(BytesIO(fut.result()), url, **conf)
                except Exception as error:  # noqa: BLE001,PERF203
                    _logger.error("Couldn't load inventory %s through handler '%s': %s", url, handler.name, error)  # noqa: TRY400
            self._inv_futures = {}

    @property
    def seen_handlers(self) -> Iterable[BaseHandler]:
        """Get the handlers that were encountered so far throughout the build.

        Returns:
            An iterable of instances of [`BaseHandler`][mkdocstrings.BaseHandler]
            (usable only to loop through it).
        """
        return self._handlers.values()

    def teardown(self) -> None:
        """Teardown all cached handlers and clear the cache."""
        for future in self._inv_futures:
            future.cancel()
        for handler in self.seen_handlers:
            handler.teardown()
        self._handlers.clear()



"""Tests for the mkdocstrings plugin."""

from __future__ import annotations

from typing import TYPE_CHECKING

from mkdocs.commands.build import build
from mkdocs.config import load_config

from mkdocstrings import MkdocstringsPlugin

if TYPE_CHECKING:
    from pathlib import Path


def test_disabling_plugin(tmp_path: Path) -> None:
    """Test disabling plugin."""
    docs_dir = tmp_path / "docs"
    site_dir = tmp_path / "site"
    docs_dir.mkdir()
    site_dir.mkdir()
    docs_dir.joinpath("index.md").write_text("::: mkdocstrings")
    config_file = tmp_path / "mkdocs.yml"
    config_file.write_text(
        """
        site_name: Test
        theme: mkdocs
        plugins:
        - mkdocstrings:
            enabled: false
        """,
    )

    mkdocs_config = load_config(str(config_file))
    mkdocs_config["docs_dir"] = str(docs_dir)
    mkdocs_config["site_dir"] = str(site_dir)
    mkdocs_config["plugins"].run_event("startup", command="build", dirty=False)
    try:
        build(mkdocs_config)
    finally:
        mkdocs_config["plugins"].run_event("shutdown")

    # make sure the instruction was not processed
    assert "::: mkdocstrings" in site_dir.joinpath("index.html").read_text(encoding="utf8")


def test_plugin_default_config(tmp_path: Path) -> None:
    """Test default config options are set for Plugin."""
    config_file_path = tmp_path / "mkdocs.yml"
    plugin = MkdocstringsPlugin()
    errors, warnings = plugin.load_config({}, config_file_path=str(config_file_path))
    assert errors == []
    assert warnings == []
    assert plugin.config == {
        "handlers": {},
        "default_handler": "python",
        "custom_templates": None,
        "enable_inventory": None,
        "enabled": True,
        "locale": None,
    }


def test_plugin_config_custom_templates(tmp_path: Path) -> None:
    """Test custom_templates option is relative to config file."""
    config_file_path = tmp_path / "mkdocs.yml"
    options = {"custom_templates": "docs/templates"}
    template_dir = tmp_path / options["custom_templates"]
    # Path must exist or config validation will fail.
    template_dir.mkdir(parents=True)
    plugin = MkdocstringsPlugin()
    errors, warnings = plugin.load_config(options, config_file_path=str(config_file_path))
    assert errors == []
    assert warnings == []
    assert plugin.config == {
        "handlers": {},
        "default_handler": "python",
        "custom_templates": str(template_dir),
        "enable_inventory": None,
        "enabled": True,
        "locale": None,
    }



"""Development tasks."""

from __future__ import annotations

import os
import re
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from duty import duty, tools

if TYPE_CHECKING:
    from duty.context import Context


PY_SRC_PATHS = (Path(_) for _ in ("src", "tests", "duties.py", "scripts"))
PY_SRC_LIST = tuple(str(_) for _ in PY_SRC_PATHS)
PY_SRC = " ".join(PY_SRC_LIST)
CI = os.environ.get("CI", "0") in {"1", "true", "yes", ""}
WINDOWS = os.name == "nt"
PTY = not WINDOWS and not CI
MULTIRUN = os.environ.get("MULTIRUN", "0") == "1"
PY_VERSION = f"{sys.version_info.major}{sys.version_info.minor}"
PY_DEV = "315"


def pyprefix(title: str) -> str:
    if MULTIRUN:
        prefix = f"(python{sys.version_info.major}.{sys.version_info.minor})"
        return f"{prefix:14}{title}"
    return title


def _get_changelog_version() -> str:
    changelog_version_re = re.compile(r"^## \[(\d+\.\d+\.\d+)\].*$")
    with Path(__file__).parent.joinpath("CHANGELOG.md").open("r", encoding="utf8") as file:
        return next(filter(bool, map(changelog_version_re.match, file))).group(1)  # ty: ignore[unresolved-attribute]


@duty
def changelog(ctx: Context, bump: str = "") -> None:
    """Update the changelog in-place with latest commits.

    Parameters:
        bump: Bump option passed to git-changelog.
    """
    ctx.run(tools.git_changelog(bump=bump or None), title="Updating changelog")
    ctx.run(tools.yore.check(bump=bump or _get_changelog_version()), title="Checking legacy code")


@duty(pre=["check-quality", "check-types", "check-docs", "check-api"])
def check(ctx: Context) -> None:
    """Check it all!"""


@duty(nofail=PY_VERSION == PY_DEV)
def check_quality(ctx: Context) -> None:
    """Check the code quality."""
    ctx.run(
        tools.ruff.check(*PY_SRC_LIST, config="config/ruff.toml", color=True),
        title=pyprefix("Checking code quality"),
    )


@duty(nofail=PY_VERSION == PY_DEV)
def check_docs(ctx: Context) -> None:
    """Check if the documentation builds correctly."""
    ctx.run(
        tools.zensical.build(strict=True),
        title=pyprefix("Building documentation"),
    )


@duty(nofail=PY_VERSION == PY_DEV)
def check_types(ctx: Context) -> None:
    """Check that the code is correctly typed."""
    py = f"{sys.version_info.major}.{sys.version_info.minor}"
    ctx.run(
        tools.ty.check(
            *PY_SRC_LIST,
            config_file="config/ty.toml",
            color=True,
            error_on_warning=True,
            python_version=py,
        ),
        title=pyprefix("Type-checking"),
    )


@duty(nofail=PY_VERSION == PY_DEV)
def check_api(ctx: Context, *cli_args: str) -> None:
    """Check for API breaking changes."""
    ctx.run(
        tools.griffe.check("mkdocstrings", search=["src"], color=True).add_args(*cli_args),
        title="Checking for API breaking changes",
        nofail=True,
    )


@duty
def docs(ctx: Context, *cli_args: str, host: str = "127.0.0.1", port: int = 8000) -> None:
    """Serve the documentation (localhost:8000).

    Parameters:
        host: The host to serve the docs from.
        port: The port to serve the docs on.
    """
    ctx.run(
        tools.zensical.serve(dev_addr=f"{host}:{port}").add_args(*cli_args),
        title="Serving documentation",
        capture=False,
    )


@duty
def docs_deploy(ctx: Context) -> None:
    """Deploy the documentation to GitHub pages."""
    from ghp_import import ghp_import  # noqa: PLC0415

    ctx.run(tools.zensical.build(), title="Building documentation site")
    ctx.run(
        ghp_import,
        kwargs={
            "srcdir": "site",
            "mesg": "chore: Update documentation",
            "push": True,
            "force": True,
            "remote": "org-pages",
        },
        title="Deploying site to GitHub Pages",
        command="ghp-import site -r org-pages -fpm 'chore: Update documentation'",
        pty=PTY,
    )


@duty
def format(ctx: Context) -> None:
    """Run formatting tools on the code."""
    ctx.run(
        tools.ruff.check(*PY_SRC_LIST, config="config/ruff.toml", fix_only=True, exit_zero=True),
        title="Auto-fixing code",
    )
    ctx.run(tools.ruff.format(*PY_SRC_LIST, config="config/ruff.toml"), title="Formatting code")


@duty
def build(ctx: Context) -> None:
    """Build source and wheel distributions."""
    ctx.run(
        ["uv", "build"],
        title="Building distributions",
        pty=PTY,
    )


@duty
def publish(ctx: Context) -> None:
    """Publish source and wheel distributions to PyPI."""
    if not Path("dist").exists():
        ctx.run("false", title="No distribution files found")
    dists = [str(dist) for dist in Path("dist").iterdir() if dist.suffix in (".gz", ".whl")]
    ctx.run(
        tools.twine.upload(*dists, skip_existing=True),
        title="Publishing distributions to PyPI",
        pty=PTY,
    )


@duty(post=["build", "publish", "docs-deploy"])
def release(ctx: Context, version: str = "") -> None:
    """Release a new Python package.

    Parameters:
        version: The new version number to use.
    """
    if not (version := (version or input("> Version to release: ")).strip()):
        ctx.run("false", title="A version must be provided")
    ctx.run("git add pyproject.toml CHANGELOG.md", title="Staging files", pty=PTY)
    ctx.run(["git", "commit", "-m", f"chore: Prepare release {version}"], title="Committing changes", pty=PTY)
    ctx.run(f"git tag -m '' -a {version}", title="Tagging commit", pty=PTY)
    ctx.run("git push", title="Pushing commits", pty=False)
    ctx.run("git push --tags", title="Pushing tags", pty=False)


@duty(silent=True, aliases=["cov"])
def coverage(ctx: Context) -> None:
    """Report coverage as text and HTML."""
    ctx.run(tools.coverage.combine(), nofail=True)
    ctx.run(tools.coverage.report(rcfile="config/coverage.ini"), capture=False)
    ctx.run(tools.coverage.html(rcfile="config/coverage.ini"))


@duty(nofail=PY_VERSION == PY_DEV)
def test(ctx: Context, *cli_args: str) -> None:
    """Run the test suite."""
    os.environ["COVERAGE_FILE"] = f".coverage.{PY_VERSION}"
    os.environ["PYTHONWARNDEFAULTENCODING"] = "1"
    ctx.run(
        tools.pytest(
            "tests",
            config_file="config/pytest.ini",
            color="yes",
        ).add_args("-n", "auto", *cli_args),
        title=pyprefix("Running tests"),
    )



"""mkdocstrings package.

Automatic documentation from sources, for MkDocs.
"""

from __future__ import annotations

from mkdocstrings._internal.extension import AutoDocProcessor, MkdocstringsExtension, makeExtension
from mkdocstrings._internal.handlers.base import (
    BaseHandler,
    CollectionError,
    CollectorItem,
    HandlerConfig,
    HandlerOptions,
    Handlers,
    ThemeNotSupported,
    do_any,
)
from mkdocstrings._internal.handlers.rendering import (
    HeadingShiftingTreeprocessor,
    Highlighter,
    IdPrependingTreeprocessor,
    MkdocstringsInnerExtension,
    ParagraphStrippingTreeprocessor,
)
from mkdocstrings._internal.inventory import Inventory, InventoryItem
from mkdocstrings._internal.loggers import (
    TEMPLATES_DIRS,
    LoggerAdapter,
    TemplateLogger,
    get_logger,
    get_template_logger,
    get_template_logger_function,
    get_template_path,
)
from mkdocstrings._internal.plugin import MkdocstringsPlugin, PluginConfig

__all__: list[str] = [
    "TEMPLATES_DIRS",
    "AutoDocProcessor",
    "BaseHandler",
    "CollectionError",
    "CollectorItem",
    "HandlerConfig",
    "HandlerOptions",
    "Handlers",
    "HeadingShiftingTreeprocessor",
    "Highlighter",
    "IdPrependingTreeprocessor",
    "Inventory",
    "InventoryItem",
    "LoggerAdapter",
    "MkdocstringsExtension",
    "MkdocstringsInnerExtension",
    "MkdocstringsPlugin",
    "ParagraphStrippingTreeprocessor",
    "PluginConfig",
    "TemplateLogger",
    "ThemeNotSupported",
    "do_any",
    "get_logger",
    "get_template_logger",
    "get_template_logger_function",
    "get_template_path",
    "makeExtension",
]



"""Tests for the internal mkdocstrings _download module."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import pytest

from mkdocstrings._internal import download

if TYPE_CHECKING:
    from collections.abc import Mapping


@pytest.mark.parametrize(
    ("credential", "expected", "env"),
    [
        ("USER", "USER", {"USER": "testuser"}),
        ("$USER", "$USER", {"USER": "testuser"}),
        ("${USER", "${USER", {"USER": "testuser"}),
        ("$USER}", "$USER}", {"USER": "testuser"}),
        ("${TOKEN}", "testtoken", {"TOKEN": "testtoken"}),
        ("${USER}:${PASSWORD}", "${USER}:testpass", {"PASSWORD": "testpass"}),
        ("${USER}:${PASSWORD}", "testuser:testpass", {"USER": "testuser", "PASSWORD": "testpass"}),
        (
            "user_prefix_${USER}_user_$uffix:pwd_prefix_${PASSWORD}_pwd_${uffix",
            "user_prefix_testuser_user_$uffix:pwd_prefix_testpass_pwd_${uffix",
            {"USER": "testuser", "PASSWORD": "testpass"},
        ),
    ],
)
def test_expand_env_vars(credential: str, expected: str, env: Mapping[str, str]) -> None:
    """Test expanding environment variables."""
    assert download._expand_env_vars(credential, url="https://test.example.com", env=env) == expected


def test_expand_env_vars_with_missing_env_var(caplog: pytest.LogCaptureFixture) -> None:
    """Test expanding environment variables with a missing environment variable."""
    caplog.set_level(logging.WARNING, logger="mkdocs.plugins.mkdocstrings._download")

    credential = "${USER}"
    env: dict[str, str] = {}
    assert download._expand_env_vars(credential, url="https://test.example.com", env=env) == "${USER}"

    output = caplog.records[0].getMessage()
    assert "'USER' is not set" in output


@pytest.mark.parametrize(
    ("url", "expected_url"),
    [
        ("http://host/path", "http://host/path"),
        ("http://token@host/path", "http://host/path"),
        ("http://${token}@host/path", "http://host/path"),
        ("http://username:password@host/path", "http://host/path"),
        ("http://username:${PASSWORD}@host/path", "http://host/path"),
        ("http://${USERNAME}:${PASSWORD}@host/path", "http://host/path"),
        ("http://prefix${USERNAME}suffix:prefix${PASSWORD}suffix@host/path", "http://host/path"),
    ],
)
def test_extract_auth_from_url(monkeypatch: pytest.MonkeyPatch, url: str, expected_url: str) -> None:
    """Test extracting the auth part from the URL."""
    monkeypatch.setattr(download, "_create_auth_header", lambda *args, **kwargs: {})
    result_url, _result_auth_header = download._extract_auth_from_url(url)
    assert result_url == expected_url


def test_create_auth_header_basic_auth() -> None:
    """Test creating the Authorization header for basic authentication."""
    auth_header = download._create_auth_header(credential="testuser:testpass", url="https://test.example.com")
    assert auth_header == {"Authorization": "Basic dGVzdHVzZXI6dGVzdHBhc3M="}


def test_create_auth_header_bearer_auth() -> None:
    """Test creating the Authorization header for bearer token authentication."""
    auth_header = download._create_auth_header(credential="token123", url="https://test.example.com")
    assert auth_header == {"Authorization": "Bearer token123"}


@pytest.mark.parametrize(
    ("var", "match"),
    [
        ("${var}", "var"),
        ("${VAR}", "VAR"),
        ("${_}", "_"),
        ("${_VAR}", "_VAR"),
        ("${VAR123}", "VAR123"),
        ("${VAR123_}", "VAR123_"),
        ("VAR", None),
        ("$1VAR", None),
        ("${1VAR}", None),
        ("${}", None),
        ("${ }", None),
    ],
)
def test_env_var_pattern(var: str, match: str | None) -> None:
    """Test the environment variable regex pattern."""
    _match = download._ENV_VAR_PATTERN.match(var)
    if _match is None:
        assert match is _match
    else:
        assert _match.group(1) == match



from __future__ import annotations

import base64
import gzip
import os
import re
import urllib.parse
import urllib.request
from typing import TYPE_CHECKING, BinaryIO

from mkdocstrings._internal.loggers import get_logger

if TYPE_CHECKING:
    from collections.abc import Mapping


_logger = get_logger("mkdocstrings")

# Regex pattern for an environment variable in the form ${ENV_VAR}.
_ENV_VAR_PATTERN = re.compile(r"\$\{([A-Za-z_][A-Za-z0-9_]*)\}")

# Timeout in seconds for downloading.
_TIMEOUT = 10


def _download_url_with_gz(url: str) -> bytes:
    url, auth_header = _extract_auth_from_url(url)

    req = urllib.request.Request(  # noqa: S310
        url,
        headers={"Accept-Encoding": "gzip", "User-Agent": "mkdocstrings/0.15.0", **auth_header},
    )
    with urllib.request.urlopen(req, timeout=_TIMEOUT) as resp:  # noqa: S310
        content: BinaryIO = resp
        if "gzip" in resp.headers.get("content-encoding", ""):
            content = gzip.GzipFile(fileobj=resp)  # ty: ignore[invalid-assignment]
        return content.read()


def _expand_env_vars(credential: str, url: str, env: Mapping[str, str] | None = None) -> str:
    """A safe implementation of environment variable substitution.

    It only supports the following forms: `${ENV_VAR}`.
    Neither `$ENV_VAR` or `%ENV_VAR` are supported.
    """
    if env is None:
        env = os.environ

    def replace_func(match: re.Match) -> str:
        try:
            return env[match.group(1)]
        except KeyError:
            _logger.warning(
                "Environment variable '%s' is not set, but is used in inventory URL %s",
                match.group(1),
                url,
            )
            return match.group(0)

    return re.sub(_ENV_VAR_PATTERN, replace_func, credential)


# Implementation adapted from PDM: https://github.com/pdm-project/pdm.
def _extract_auth_from_url(url: str) -> tuple[str, dict[str, str]]:
    """Extract credentials from the URL if present, and return the URL and the appropriate auth header for the credentials."""
    if "@" not in url:
        return url, {}

    scheme, netloc, *rest = urllib.parse.urlparse(url)
    auth, host = netloc.split("@", 1)
    auth = _expand_env_vars(credential=auth, url=url)
    auth_header = _create_auth_header(credential=auth, url=url)

    url = urllib.parse.urlunparse((scheme, host, *rest))
    return url, auth_header


def _create_auth_header(credential: str, url: str) -> dict[str, str]:
    """Create the Authorization header for basic or bearer authentication, depending on credential."""
    if ":" not in credential:
        # We assume that the user is using a token.
        _logger.debug("Using bearer token authentication for %s", url)
        return {"Authorization": f"Bearer {credential}"}

    # Else, we assume that the user is using user:password.
    user, pwd = credential.split(":", 1)
    _logger.debug("Using basic authentication for %s", url)
    credentials = base64.encodebytes(f"{user}:{pwd}".encode()).decode().strip()
    return {"Authorization": f"Basic {credentials}"}



# Module responsible for the objects inventory.
#
# Credits to Brian Skinn and the sphobjinv project:
# https://github.com/bskinn/sphobjinv.

from __future__ import annotations

import re
import zlib
from textwrap import dedent
from typing import TYPE_CHECKING, BinaryIO, Literal, overload

if TYPE_CHECKING:
    from collections.abc import Collection


class InventoryItem:
    """Inventory item."""

    def __init__(
        self,
        name: str,
        domain: str,
        role: str,
        uri: str,
        priority: int = 1,
        dispname: str | None = None,
    ):
        """Initialize the object.

        Arguments:
            name: The item name.
            domain: The item domain, like 'python' or 'crystal'.
            role: The item role, like 'class' or 'method'.
            uri: The item URI.
            priority: The item priority. Only used internally by mkdocstrings and Sphinx.
            dispname: The item display name.
        """
        self.name: str = name
        """The item name."""
        self.domain: str = domain
        """The item domain."""
        self.role: str = role
        """The item role."""
        self.uri: str = uri
        """The item URI."""
        self.priority: int = priority
        """The item priority."""
        self.dispname: str = dispname or name
        """The item display name."""

    def format_sphinx(self) -> str:
        """Format this item as a Sphinx inventory line.

        Returns:
            A line formatted for an `objects.inv` file.
        """
        dispname = self.dispname
        if dispname == self.name:
            dispname = "-"
        uri = self.uri
        if uri.endswith(self.name):
            uri = uri[: -len(self.name)] + "$"
        return f"{self.name} {self.domain}:{self.role} {self.priority} {uri} {dispname}"

    sphinx_item_regex = re.compile(r"^(.+?)\s+(\S+):(\S+)\s+(-?\d+)\s+(\S+)\s*(.*)$")
    """Regex to parse a Sphinx v2 inventory line."""

    @overload
    @classmethod
    def parse_sphinx(cls, line: str, *, return_none: Literal[False]) -> InventoryItem: ...

    @overload
    @classmethod
    def parse_sphinx(cls, line: str, *, return_none: Literal[True]) -> InventoryItem | None: ...

    @classmethod
    def parse_sphinx(cls, line: str, *, return_none: bool = False) -> InventoryItem | None:
        """Parse a line from a Sphinx v2 inventory file and return an `InventoryItem` from it."""
        match = cls.sphinx_item_regex.search(line)
        if not match:
            if return_none:
                return None
            raise ValueError(line)
        name, domain, role, priority, uri, dispname = match.groups()
        if uri.endswith("$"):
            uri = uri[:-1] + name
        if dispname == "-":
            dispname = name
        return cls(name, domain, role, uri, int(priority), dispname)


class Inventory(dict):
    """Inventory of collected and rendered objects."""

    def __init__(self, items: list[InventoryItem] | None = None, project: str = "project", version: str = "0.0.0"):
        """Initialize the object.

        Arguments:
            items: A list of items.
            project: The project name.
            version: The project version.
        """
        super().__init__()
        items = items or []
        for item in items:
            self[item.name] = item
        self.project = project
        """The project name."""
        self.version = version
        """The project version."""

    def register(
        self,
        name: str,
        domain: str,
        role: str,
        uri: str,
        priority: int = 1,
        dispname: str | None = None,
    ) -> None:
        """Create and register an item.

        Arguments:
            name: The item name.
            domain: The item domain, like 'python' or 'crystal'.
            role: The item role, like 'class' or 'method'.
            uri: The item URI.
            priority: The item priority. Only used internally by mkdocstrings and Sphinx.
            dispname: The item display name.
        """
        self[name] = InventoryItem(
            name=name,
            domain=domain,
            role=role,
            uri=uri,
            priority=priority,
            dispname=dispname,
        )

    def format_sphinx(self) -> bytes:
        """Format this inventory as a Sphinx `objects.inv` file.

        Returns:
            The inventory as bytes.
        """
        header = (
            dedent(
                f"""
                # Sphinx inventory version 2
                # Project: {self.project}
                # Version: {self.version}
                # The remainder of this file is compressed using zlib.
                """,
            )
            .lstrip()
            .encode("utf8")
        )

        lines = [
            item.format_sphinx().encode("utf8")
            for item in sorted(self.values(), key=lambda item: (item.domain, item.name))
        ]
        return header + zlib.compress(b"\n".join(lines) + b"\n", 9)

    @classmethod
    def parse_sphinx(cls, in_file: BinaryIO, *, domain_filter: Collection[str] = ()) -> Inventory:
        """Parse a Sphinx v2 inventory file and return an `Inventory` from it.

        Arguments:
            in_file: The binary file-like object to read from.
            domain_filter: A collection of domain values to allow (and filter out all other ones).

        Returns:
            An inventory containing the collected items.
        """
        in_file.readline()  # "Sphinx inventory version X"
        project = in_file.readline().removeprefix(b"# Project: ").decode("utf8").rstrip()
        version = in_file.readline().removeprefix(b"# Version: ").decode("utf8").rstrip()
        in_file.readline()  # "The remainder of this file is compressed using zlib"

        lines = zlib.decompress(in_file.read()).splitlines()
        items: list[InventoryItem] = [
            item for line in lines if (item := InventoryItem.parse_sphinx(line.decode("utf8"), return_none=True))
        ]
        if domain_filter:
            items = [item for item in items if item.domain in domain_filter]
        return cls(items, project=project, version=version)



# This module contains the "mkdocstrings" plugin for MkDocs.
#
# The plugin instantiates a Markdown extension ([`MkdocstringsExtension`][mkdocstrings.MkdocstringsExtension]),
# and adds it to the list of Markdown extensions used by `mkdocs`
# during the [`on_config` event hook](https://www.mkdocs.org/user-guide/plugins/#on_config).
#
# Once the documentation is built, the [`on_post_build` event hook](https://www.mkdocs.org/user-guide/plugins/#on_post_build)
# is triggered and calls the [`handlers.teardown()` method][mkdocstrings.Handlers.teardown]. This method is
# used to teardown the handlers that were instantiated during documentation buildup.
#
# Finally, when serving the documentation, it can add directories to watch
# during the [`on_serve` event hook](https://www.mkdocs.org/user-guide/plugins/#on_serve).

from __future__ import annotations

import os
import re
from functools import partial
from inspect import signature
from re import Match
from typing import TYPE_CHECKING, Any

from mkdocs.config import Config
from mkdocs.config import config_options as opt
from mkdocs.plugins import BasePlugin, CombinedEvent, event_priority
from mkdocs.utils import write_file
from mkdocs_autorefs import AutorefsConfig, AutorefsPlugin

from mkdocstrings._internal.extension import MkdocstringsExtension
from mkdocstrings._internal.handlers.base import BaseHandler, Handlers
from mkdocstrings._internal.loggers import get_logger

if TYPE_CHECKING:
    from jinja2.environment import Environment
    from mkdocs.config.defaults import MkDocsConfig
    from mkdocs.structure.files import Files


_logger = get_logger("mkdocstrings")


class PluginConfig(Config):
    """The configuration options of `mkdocstrings`, written in `mkdocs.yml`."""

    handlers = opt.Type(dict, default={})
    """
    Global configuration of handlers.

    You can set global configuration per handler, applied everywhere,
    but overridable in each "autodoc" instruction. Example:

    ```yaml
    plugins:
      - mkdocstrings:
          handlers:
            python:
              options:
                option1: true
                option2: "value"
            rust:
              options:
                option9: 2
    ```
    """

    default_handler = opt.Type(str, default="python")
    """The default handler to use. The value is the name of the handler module. Default is "python"."""
    custom_templates = opt.Optional(opt.Dir(exists=True))
    """Location of custom templates to use when rendering API objects.

    Value should be the path of a directory relative to the MkDocs configuration file.
    """
    enable_inventory = opt.Optional(opt.Type(bool))
    """Whether to enable object inventory creation."""
    enabled = opt.Type(bool, default=True)
    """Whether to enable the plugin. Default is true. If false, *mkdocstrings* will not collect or render anything."""
    locale = opt.Optional(opt.Type(str))
    """The locale to use for translations."""


class MkdocstringsPlugin(BasePlugin[PluginConfig]):
    """An `mkdocs` plugin.

    This plugin defines the following event hooks:

    - `on_config`
    - `on_env`
    - `on_post_build`

    Check the [Developing Plugins](https://www.mkdocs.org/user-guide/plugins/#developing-plugins) page of `mkdocs`
    for more information about its plugin system.
    """

    css_filename: str = "assets/_mkdocstrings.css"
    """The path of the CSS file to write in the site directory."""

    def __init__(self) -> None:
        """Initialize the object."""
        super().__init__()
        self._handlers: Handlers | None = None

    @property
    def handlers(self) -> Handlers:
        """Get the instance of [mkdocstrings.Handlers][] for this plugin/build.

        Raises:
            RuntimeError: If the plugin hasn't been initialized with a config.

        Returns:
            An instance of [mkdocstrings.Handlers][] (the same throughout the build).
        """
        if not self._handlers:
            raise RuntimeError("The plugin hasn't been initialized with a config yet")
        return self._handlers

    def on_config(self, config: MkDocsConfig) -> MkDocsConfig | None:
        """Instantiate our Markdown extension.

        Hook for the [`on_config` event](https://www.mkdocs.org/user-guide/plugins/#on_config).
        In this hook, we instantiate our [`MkdocstringsExtension`][mkdocstrings.MkdocstringsExtension]
        and add it to the list of Markdown extensions used by `mkdocs`.

        We pass this plugin's configuration dictionary to the extension when instantiating it (it will need it
        later when processing markdown to get handlers and their global configurations).

        Arguments:
            config: The MkDocs config object.

        Returns:
            The modified config.
        """
        if not self.plugin_enabled:
            _logger.debug("Plugin is not enabled. Skipping.")
            return config
        _logger.debug("Adding extension to the list")

        locale = self.config.locale or config.theme.get("language") or config.theme.get("locale") or "en"
        locale = str(locale).replace("_", "-")

        handlers = Handlers(
            default=self.config.default_handler,
            handlers_config=self.config.handlers,
            theme=config.theme.name or os.path.dirname(config.theme.dirs[0]),  # noqa: PTH120
            custom_templates=self.config.custom_templates,
            mdx=config.markdown_extensions,
            mdx_config=config.mdx_configs,
            inventory_project=config.site_name,
            inventory_version="0.0.0",  # TODO: Find a way to get actual version.
            locale=locale,
            tool_config=config,
        )

        handlers._download_inventories()

        AutorefsPlugin.record_backlinks = True
        autorefs: AutorefsPlugin
        try:
            # If autorefs plugin is explicitly enabled, just use it.
            autorefs = config.plugins["autorefs"]  # ty: ignore[invalid-assignment]
            _logger.debug("Picked up existing autorefs instance %r", autorefs)
        except KeyError:
            # Otherwise, add a limited instance of it that acts only on what's added through `register_anchor`.
            autorefs = AutorefsPlugin()
            autorefs.config = AutorefsConfig()  # ty:ignore[invalid-assignment]
            autorefs.scan_toc = False
            config.plugins["autorefs"] = autorefs
            _logger.debug("Added a subdued autorefs instance %r", autorefs)

        mkdocstrings_extension = MkdocstringsExtension(handlers, autorefs)
        config.markdown_extensions.append(mkdocstrings_extension)  # ty: ignore[invalid-argument-type]

        config.extra_css.insert(0, self.css_filename)  # So that it has lower priority than user files.

        self._autorefs = autorefs
        self._handlers = handlers
        return config

    @property
    def inventory_enabled(self) -> bool:
        """Tell if the inventory is enabled or not.

        Returns:
            Whether the inventory is enabled.
        """
        inventory_enabled = self.config.enable_inventory
        if inventory_enabled is None:
            inventory_enabled = any(handler.enable_inventory for handler in self.handlers.seen_handlers)
        return inventory_enabled

    @property
    def plugin_enabled(self) -> bool:
        """Tell if the plugin is enabled or not.

        Returns:
            Whether the plugin is enabled.
        """
        return self.config.enabled

    @event_priority(50)  # Early, before autorefs' starts applying cross-refs and collecting backlinks.
    def _on_env_load_inventories(self, env: Environment, config: MkDocsConfig, *args: Any, **kwargs: Any) -> None:  # noqa: ARG002
        if self.plugin_enabled and self._handlers:
            register = config.plugins["autorefs"].register_url  # ty: ignore[unresolved-attribute]
            for identifier, url in self._handlers._yield_inventory_items():
                register(identifier, url)

    @event_priority(-20)  # Late, not important.
    def _on_env_add_css(self, env: Environment, config: MkDocsConfig, *args: Any, **kwargs: Any) -> None:  # noqa: ARG002
        if self.plugin_enabled and self._handlers:
            css_content = "\n".join(handler.extra_css for handler in self.handlers.seen_handlers)
            write_file(css_content.encode("utf-8"), os.path.join(config.site_dir, self.css_filename))  # noqa: PTH118

    @event_priority(-20)  # Late, not important.
    def _on_env_write_inventory(self, env: Environment, config: MkDocsConfig, *args: Any, **kwargs: Any) -> None:  # noqa: ARG002
        if self.plugin_enabled and self._handlers and self.inventory_enabled:
            _logger.debug("Creating inventory file objects.inv")
            inv_contents = self.handlers.inventory.format_sphinx()
            write_file(inv_contents, os.path.join(config.site_dir, "objects.inv"))  # noqa: PTH118

    @event_priority(-100)  # Last, after autorefs has finished applying cross-refs and collecting backlinks.
    def _on_env_apply_backlinks(self, env: Environment, /, *, config: MkDocsConfig, files: Files) -> Environment:  # noqa: ARG002
        regex = re.compile(r"<backlinks\s+identifier=\"([^\"]+)\"\s+handler=\"([^\"]+)\"\s*/?>")

        def repl(match: Match) -> str:
            handler_name = match.group(2)
            handler = self.handlers.get_handler(handler_name)

            # The handler doesn't implement backlinks,
            # return early to avoid computing them.
            if handler.render_backlinks.__func__ is BaseHandler.render_backlinks:
                return ""

            identifier = match.group(1)
            aliases = handler.get_aliases(identifier)
            backlinks = self._autorefs.get_backlinks(identifier, *aliases, from_url=file.page.url)

            # No backlinks, avoid calling the handler's method.
            if not backlinks:
                return ""

            if "locale" in signature(handler.render_backlinks).parameters:
                render_backlinks = partial(handler.render_backlinks, locale=self.handlers._locale)
            else:
                render_backlinks = handler.render_backlinks

            return render_backlinks(backlinks)

        for file in files:
            if file.page and file.page.content:
                _logger.debug("Applying backlinks in page %s", file.page.file.src_path)
                file.page.content = regex.sub(repl, file.page.content)

        return env

    on_env = CombinedEvent(_on_env_load_inventories, _on_env_add_css, _on_env_write_inventory, _on_env_apply_backlinks)
    """Extra actions that need to happen after all Markdown-to-HTML page rendering.

    Hook for the [`on_env` event](https://www.mkdocs.org/user-guide/plugins/#on_env).

    - Gather results from background inventory download tasks.
    - Write mkdocstrings' extra files (CSS, inventory) into the site directory.
    - Apply backlinks to the HTML output of each page.
    """

    def on_post_build(
        self,
        config: MkDocsConfig,  # noqa: ARG002
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        """Teardown the handlers.

        Hook for the [`on_post_build` event](https://www.mkdocs.org/user-guide/plugins/#on_post_build).
        This hook is used to teardown all the handlers that were instantiated and cached during documentation buildup.

        For example, a handler could open a subprocess in the background and keep it open
        to feed it "autodoc" instructions and get back JSON data. If so, it should then close the subprocess at some point:
        the proper place to do this is in the handler's `teardown` method, which is indirectly called by this hook.

        Arguments:
            config: The MkDocs config object.
            **kwargs: Additional arguments passed by MkDocs.
        """
        if not self.plugin_enabled:
            return

        if self._handlers:
            _logger.debug("Tearing handlers down")
            self.handlers.teardown()

    def get_handler(self, handler_name: str) -> BaseHandler:
        """Get a handler by its name. See [mkdocstrings.Handlers.get_handler][].

        Arguments:
            handler_name: The name of the handler.

        Returns:
            An instance of a subclass of [`BaseHandler`][mkdocstrings.BaseHandler].
        """
        return self.handlers.get_handler(handler_name)



# This module holds helpers responsible for augmentations to the Markdown sub-documents produced by handlers.

from __future__ import annotations

import copy
import re
import textwrap
from typing import TYPE_CHECKING, Any

from markdown.extensions import Extension
from markdown.extensions.codehilite import CodeHiliteExtension
from markdown.treeprocessors import Treeprocessor
from markupsafe import Markup
from pymdownx.highlight import Highlight, HighlightExtension

if TYPE_CHECKING:
    from xml.etree.ElementTree import Element

    from markdown import Markdown


class Highlighter(Highlight):
    """Code highlighter that tries to match the Markdown configuration.

    Picking up the global config and defaults works only if you use the `codehilite` or
    `pymdownx.highlight` (recommended) Markdown extension.

    -   If you use `pymdownx.highlight`, highlighting settings are picked up from it, and the
        default CSS class is `.highlight`. This also means the default of `guess_lang: false`.

    -   Otherwise, if you use the `codehilite` extension, settings are picked up from it, and the
        default CSS class is `.codehilite`. Also consider setting `guess_lang: false`.

    -   If neither are added to `markdown_extensions`, highlighting is enabled anyway. This is for
        backwards compatibility. If you really want to disable highlighting even in *mkdocstrings*,
        add one of these extensions anyway and set `use_pygments: false`.

    The underlying implementation is `pymdownx.highlight` regardless.
    """

    # https://raw.githubusercontent.com/facelessuser/pymdown-extensions/main/docs/src/markdown/extensions/highlight.md
    _highlight_config_keys = frozenset(
        (
            "css_class",
            "guess_lang",
            "default_lang",
            "pygments_style",
            "noclasses",
            "use_pygments",
            "linenums",
            "linenums_special",
            "linenums_style",
            "linenums_class",
            "extend_pygments_lang",
            "language_prefix",
            "code_attr_on_pre",
            "auto_title",
            "auto_title_map",
            "line_spans",
            "anchor_linenums",
            "line_anchors",
            "pygments_lang_class",
            "stripnl",
        ),
    )

    def __init__(self, md: Markdown):
        """Configure to match a `markdown.Markdown` instance.

        Arguments:
            md: The Markdown instance to read configs from.
        """
        config: dict[str, Any] = {}
        self._highlighter: str | None = None
        for ext in md.registeredExtensions:
            if isinstance(ext, HighlightExtension) and (ext.enabled or not config):
                self._highlighter = "highlight"
                config = ext.getConfigs()
                break  # This one takes priority, no need to continue looking
            if isinstance(ext, CodeHiliteExtension) and not config:
                self._highlighter = "codehilite"
                config = ext.getConfigs()
                config["language_prefix"] = config["lang_prefix"]
        self._css_class = config.pop("css_class", "highlight")
        super().__init__(**{name: opt for name, opt in config.items() if name in self._highlight_config_keys})

    def highlight(  # ty: ignore[invalid-method-override]
        self,
        src: str,
        language: str | None = None,
        *,
        inline: bool = False,
        dedent: bool = True,
        linenums: bool | None = None,
        **kwargs: Any,
    ) -> str:
        """Highlight a code-snippet.

        Arguments:
            src: The code to highlight.
            language: Explicitly tell what language to use for highlighting.
            inline: Whether to highlight as inline.
            dedent: Whether to dedent the code before highlighting it or not.
            linenums: Whether to add line numbers in the result.
            **kwargs: Pass on to `pymdownx.highlight.Highlight.highlight`.

        Returns:
            The highlighted code as HTML text, marked safe (not escaped for HTML).
        """
        if isinstance(src, Markup):
            src = src.unescape()
        if dedent:
            src = textwrap.dedent(src)

        kwargs.setdefault("css_class", self._css_class)
        old_linenums = self.linenums
        if linenums is not None:
            self.linenums = linenums
        try:
            result = super().highlight(src, language, inline=inline, **kwargs)
        finally:
            self.linenums = old_linenums

        if inline:
            # From the maintainer of codehilite, the codehilite CSS class, as defined by the user,
            # should never be added to inline code, because codehilite does not support inline code.
            # See https://github.com/Python-Markdown/markdown/issues/1220#issuecomment-1692160297.
            css_class = "" if self._highlighter == "codehilite" else kwargs["css_class"]
            return Markup(f'<code class="{css_class} language-{language}">{result.text}</code>')
        return Markup(result)


class IdPrependingTreeprocessor(Treeprocessor):
    """Prepend the configured prefix to IDs of all HTML elements."""

    name: str = "mkdocstrings_ids"
    """The name of the treeprocessor."""

    id_prefix: str
    """The prefix to add to every ID. It is prepended without any separator; specify your own separator if needed."""

    def __init__(self, md: Markdown, id_prefix: str):
        """Initialize the object.

        Arguments:
            md: A `markdown.Markdown` instance.
            id_prefix: The prefix to add to every ID. It is prepended without any separator.
        """
        super().__init__(md)
        self.id_prefix = id_prefix

    def run(self, root: Element) -> None:
        """Prepend the configured prefix to all IDs in the document."""
        if self.id_prefix:
            self._prefix_ids(root)

    def _prefix_ids(self, root: Element) -> None:
        index = len(root)
        for el in reversed(root):  # Reversed mainly for the ability to mutate during iteration.
            index -= 1

            self._prefix_ids(el)
            href_attr = el.get("href")

            if id_attr := el.get("id"):
                if el.tag == "a" and not href_attr:
                    # An anchor with id and no href is used by autorefs:
                    # leave it untouched and insert a copy with updated id after it.
                    new_el = copy.deepcopy(el)
                    new_el.set("id", self.id_prefix + id_attr)
                    root.insert(index + 1, new_el)
                else:
                    # Anchors with id and href are not used by autorefs:
                    # update in place.
                    el.set("id", self.id_prefix + id_attr)

            # Always update hrefs, names and labels-for:
            # there will always be a corresponding id.
            if href_attr and href_attr.startswith("#"):
                el.set("href", "#" + self.id_prefix + href_attr[1:])

            if name_attr := el.get("name"):
                el.set("name", self.id_prefix + name_attr)

            if el.tag == "label":
                for_attr = el.get("for")
                if for_attr:
                    el.set("for", self.id_prefix + for_attr)


class HeadingShiftingTreeprocessor(Treeprocessor):
    """Shift levels of all Markdown headings according to the configured base level."""

    name: str = "mkdocstrings_headings"
    """The name of the treeprocessor."""

    regex: re.Pattern = re.compile(r"([Hh])([1-6])")
    """The regex to match heading tags."""

    shift_by: int
    """The number of heading "levels" to add to every heading. `<h2>` with `shift_by = 3` becomes `<h5>`."""

    def __init__(self, md: Markdown, shift_by: int):
        """Initialize the object.

        Arguments:
            md: A `markdown.Markdown` instance.
            shift_by: The number of heading "levels" to add to every heading.
        """
        super().__init__(md)
        self.shift_by = shift_by

    def run(self, root: Element) -> None:
        """Shift the levels of all headings in the document."""
        if not self.shift_by:
            return
        for el in root.iter():
            match = self.regex.fullmatch(el.tag)
            if match:
                level = int(match[2]) + self.shift_by
                level = max(1, min(level, 6))
                el.tag = f"{match[1]}{level}"


class _HeadingReportingTreeprocessor(Treeprocessor):
    """Records the heading elements encountered in the document."""

    name: str = "mkdocstrings_headings_list"
    """The name of the treeprocessor."""

    regex: re.Pattern = re.compile(r"[Hh][1-6]")
    """The regex to match heading tags."""

    headings: list[Element]
    """The list (the one passed in the initializer) that is used to record the heading elements (by appending to it)."""

    def __init__(self, md: Markdown, headings: list[Element]):
        super().__init__(md)
        self.headings = headings

    def run(self, root: Element) -> None:
        """Record all heading elements encountered in the document."""
        permalink_class = self.md.treeprocessors["toc"].permalink_class
        for el in root.iter():
            if self.regex.fullmatch(el.tag):
                el = copy.copy(el)  # noqa: PLW2901
                # 'toc' extension's first pass (which we require to build heading stubs/ids) also edits the HTML.
                # Undo the permalink edit so we can pass this heading to the outer pass of the 'toc' extension.
                if len(el) > 0 and el[-1].get("class") == permalink_class:
                    del el[-1]
                self.headings.append(el)


class ParagraphStrippingTreeprocessor(Treeprocessor):
    """Unwraps the `<p>` element around the whole output."""

    name: str = "mkdocstrings_strip_paragraph"
    """The name of the treeprocessor."""

    strip: bool = False
    """Whether to strip `<p>` elements or not."""

    def run(self, root: Element) -> Element | None:
        """Unwrap the root element if it's a single `<p>` element."""
        if self.strip and len(root) == 1 and root[0].tag == "p":
            # Turn the single `<p>` element into the root element and inherit its tag name (it's significant!)
            root[0].tag = root.tag
            return root[0]
        return None


class MkdocstringsInnerExtension(Extension):
    """Extension that should always be added to Markdown sub-documents that handlers request (and *only* them)."""

    def __init__(self, headings: list[Element]):
        """Initialize the object.

        Arguments:
            headings: A list that will be populated with all HTML heading elements encountered in the document.
        """
        super().__init__()
        self.headings = headings
        """The list that will be populated with all HTML heading elements encountered in the document."""

    def extendMarkdown(self, md: Markdown) -> None:  # noqa: N802 (casing: parent method's name)
        """Register the extension.

        Arguments:
            md: A `markdown.Markdown` instance.
        """
        md.registerExtension(self)
        md.treeprocessors.register(
            HeadingShiftingTreeprocessor(md, 0),
            HeadingShiftingTreeprocessor.name,
            priority=12,
        )
        md.treeprocessors.register(
            IdPrependingTreeprocessor(md, ""),
            IdPrependingTreeprocessor.name,
            priority=4,  # Right after 'toc' (needed because that extension adds ids to headers).
        )
        md.treeprocessors.register(
            _HeadingReportingTreeprocessor(md, self.headings),
            _HeadingReportingTreeprocessor.name,
            priority=1,  # Close to the end.
        )
        md.treeprocessors.register(
            ParagraphStrippingTreeprocessor(md),
            ParagraphStrippingTreeprocessor.name,
            priority=0.99,  # Close to the end.
        )



# This module holds the code of the Markdown extension responsible for matching "autodoc" instructions.
#
# The extension is composed of a Markdown [block processor](https://python-markdown.github.io/extensions/api/#blockparser)
# that matches indented blocks starting with a line like `::: identifier`.
#
# For each of these blocks, it uses a [handler][mkdocstrings.BaseHandler] to collect documentation about
# the given identifier and render it with Jinja templates.
#
# Both the collection and rendering process can be configured by adding YAML configuration under the "autodoc"
# instruction:
#
# ```yaml
# ::: some.identifier
#     handler: python
#     options:
#       option1: value1
#       option2:
#       - value2a
#       - value2b
#       option_x: etc
# ```

from __future__ import annotations

import re
from functools import partial
from inspect import signature
from typing import TYPE_CHECKING, Any
from xml.etree.ElementTree import Element

import yaml
from jinja2.exceptions import TemplateNotFound
from markdown.blockprocessors import BlockProcessor
from markdown.extensions import Extension
from markdown.treeprocessors import Treeprocessor
from mkdocs.exceptions import PluginError
from mkdocs_autorefs import AutorefsConfig, AutorefsExtension, AutorefsPlugin

from mkdocstrings._internal.handlers.base import BaseHandler, CollectionError, CollectorItem, Handlers
from mkdocstrings._internal.loggers import get_logger

if TYPE_CHECKING:
    from collections.abc import MutableSequence

    from markdown import Markdown


_logger = get_logger("mkdocstrings")


class AutoDocProcessor(BlockProcessor):
    """Our "autodoc" Markdown block processor.

    It has a [`test` method][mkdocstrings.AutoDocProcessor.test] that tells if a block matches a criterion,
    and a [`run` method][mkdocstrings.AutoDocProcessor.run] that processes it.

    It also has utility methods allowing to get handlers and their configuration easily, useful when processing
    a matched block.
    """

    regex = re.compile(r"^(?P<heading>#{1,6} *|)::: ?(?P<name>.+?) *$", flags=re.MULTILINE)
    """The regular expression to match our autodoc instructions."""

    def __init__(
        self,
        md: Markdown,
        *,
        handlers: Handlers,
        autorefs: AutorefsPlugin,
    ) -> None:
        """Initialize the object.

        Arguments:
            md: A `markdown.Markdown` instance.
            handlers: The handlers container.
            autorefs: The autorefs plugin instance.
        """
        super().__init__(parser=md.parser)
        self.md = md
        """The Markdown instance."""
        self._handlers = handlers
        self._autorefs = autorefs
        self._updated_envs: set = set()

    def test(self, parent: Element, block: str) -> bool:  # noqa: ARG002
        """Match our autodoc instructions.

        Arguments:
            parent: The parent element in the XML tree.
            block: The block to be tested.

        Returns:
            Whether this block should be processed or not.
        """
        return bool(self.regex.search(block))

    def run(self, parent: Element, blocks: MutableSequence[str]) -> None:
        """Run code on the matched blocks.

        The identifier and configuration lines are retrieved from a matched block
        and used to collect and render an object.

        Arguments:
            parent: The parent element in the XML tree.
            blocks: The rest of the blocks to be processed.
        """
        block = blocks.pop(0)
        match = self.regex.search(block)

        if match:
            if match.start() > 0:
                self.parser.parseBlocks(parent, [block[: match.start()]])
            # removes the first line
            block = block[match.end() :]

        block, the_rest = self.detab(block)

        if not block and blocks and blocks[0].startswith(("    handler:", "    options:")):
            # YAML options were separated from the `:::` line by a blank line.
            block = blocks.pop(0)

        if match:
            identifier = match["name"]
            heading_level = match["heading"].count("#")
            _logger.debug("Matched '::: %s'", identifier)

            html, handler, _ = self._process_block(identifier, block, heading_level)
            el = Element("div", {"class": "mkdocstrings"})
            # The final HTML is inserted as opaque to subsequent processing, and only revealed at the end.
            el.text = self.md.htmlStash.store(html)

            if handler.outer_layer:
                self._process_headings(handler, el)

            parent.append(el)

        if the_rest:
            # This block contained unindented line(s) after the first indented
            # line. Insert these lines as the first block of the master blocks
            # list for future processing.
            blocks.insert(0, the_rest)

    def _process_block(
        self,
        identifier: str,
        yaml_block: str,
        heading_level: int = 0,
    ) -> tuple[str, BaseHandler, CollectorItem]:
        """Process an autodoc block.

        Arguments:
            identifier: The identifier of the object to collect and render.
            yaml_block: The YAML configuration.
            heading_level: Suggested level of the heading to insert (0 to ignore).

        Raises:
            PluginError: When something wrong happened during collection.
            TemplateNotFound: When a template used for rendering could not be found.

        Returns:
            Rendered HTML, the handler that was used, and the collected item.
        """
        local_config = yaml.safe_load(yaml_block) or {}
        handler_name = self._handlers.get_handler_name(local_config)

        _logger.debug("Using handler '%s'", handler_name)
        handler = self._handlers.get_handler(handler_name)

        local_options = local_config.get("options", {})
        if heading_level:
            # Heading level obtained from Markdown (`##`) takes precedence.
            local_options["heading_level"] = heading_level

        options = handler.get_options(local_options)

        _logger.debug("Collecting data")
        try:
            data: CollectorItem = handler.collect(identifier, options)
        except CollectionError as exception:
            _logger.error("%s", exception)  # noqa: TRY400
            raise PluginError(f"Could not collect '{identifier}'") from exception

        if handler_name not in self._updated_envs:  # We haven't seen this handler before on this document.
            _logger.debug("Updating handler's rendering env")
            handler._update_env(self.md, config=self._handlers._tool_config)
            self._updated_envs.add(handler_name)

        _logger.debug("Rendering templates")
        if "locale" in signature(handler.render).parameters:
            render = partial(handler.render, locale=self._handlers._locale)
        else:
            render = handler.render
        try:
            rendered = render(data, options)
        except TemplateNotFound as exc:
            _logger.error(  # noqa: TRY400
                "Template '%s' not found for '%s' handler and theme '%s'.",
                exc.name,
                handler_name,
                self._handlers._theme,
            )
            raise

        return rendered, handler, data

    def _process_headings(self, handler: BaseHandler, element: Element) -> None:
        # We're in the outer handler layer, as well as the outer extension layer.
        #
        # The "handler layer" tracks the nesting of the autodoc blocks, which can appear in docstrings.
        #
        # - Render ::: Object1                  # Outer handler layer
        #   - Render Object1's docstring        # Outer handler layer
        #     - Docstring renders ::: Object2   # Inner handler layers
        #       - etc.                          # Inner handler layers
        #
        # The "extension layer" tracks whether we're converting an autodoc instruction
        # or nested content within it, like docstrings. Markdown conversion within Markdown conversion.
        #
        # - Render ::: Object1                  # Outer extension layer
        #   - Render Object1's docstring        # Inner extension layer
        #
        # The generated HTML was just stashed, and the `toc` extension won't be able to see headings.
        # We need to duplicate the headings directly, just so `toc` can pick them up,
        # otherwise they wouldn't appear in the final table of contents.
        #
        # These headings are generated by the `BaseHandler.do_heading` method (Jinja filter),
        # which runs in the inner extension layer, and not in the outer one where we are now.
        headings = handler.get_headings()
        element.extend(headings)
        # These duplicated headings will later be removed by our `_HeadingsPostProcessor` processor,
        # which runs right after `toc` (see `MkdocstringsExtension.extendMarkdown`).
        #
        # If we were in an inner handler layer, we wouldn't do any of this
        # and would just let headings bubble up to the outer handler layer.

        if (page := self._autorefs.current_page) is None:
            return

        for heading in headings:
            rendered_id = heading.attrib["id"]

            skip_inventory = "data-skip-inventory" in heading.attrib
            if skip_inventory:
                _logger.debug(
                    "Skipping heading with id %r because data-skip-inventory is present",
                    rendered_id,
                )
                continue

            # The title is registered to be used as tooltip by autorefs.
            self._autorefs.register_anchor(page, rendered_id, title=heading.text, primary=True)

            # Register all identifiers for this object
            # both in the autorefs plugin and in the inventory.
            aliases: tuple[str, ...]
            aliases = handler.get_aliases(rendered_id)

            for alias in aliases:
                if alias != rendered_id:
                    self._autorefs.register_anchor(page, alias, rendered_id, primary=False)

            if "data-role" in heading.attrib:
                self._handlers.inventory.register(
                    name=rendered_id,
                    domain=handler.domain,
                    role=heading.attrib["data-role"],
                    priority=1,  # Register with standard priority.
                    uri=f"{page.url}#{rendered_id}",
                )
                for alias in aliases:
                    if alias not in self._handlers.inventory:
                        self._handlers.inventory.register(
                            name=alias,
                            domain=handler.domain,
                            role=heading.attrib["data-role"],
                            priority=2,  # Register with lower priority.
                            uri=f"{page.url}#{rendered_id}",
                        )


class _HeadingsPostProcessor(Treeprocessor):
    def run(self, root: Element) -> None:
        self._remove_duplicated_headings(root)

    def _remove_duplicated_headings(self, parent: Element) -> None:
        carry_text = ""
        for el in reversed(parent):  # Reversed mainly for the ability to mutate during iteration.
            if el.tag == "div" and el.get("class") == "mkdocstrings":
                # Delete the duplicated headings along with their container, but keep the text (i.e. the actual HTML).
                carry_text = (el.text or "") + carry_text
                parent.remove(el)
            else:
                if carry_text:
                    el.tail = (el.tail or "") + carry_text
                    carry_text = ""
                self._remove_duplicated_headings(el)

        if carry_text:
            parent.text = (parent.text or "") + carry_text


class _TocLabelsTreeProcessor(Treeprocessor):
    def run(self, root: Element) -> None:  # noqa: ARG002
        self._override_toc_labels(self.md.toc_tokens)  # ty: ignore[unresolved-attribute]

    def _override_toc_labels(self, tokens: list[dict[str, Any]]) -> None:
        for token in tokens:
            if (label := token.get("data-toc-label")) and token["name"] != label:
                token["name"] = label
            self._override_toc_labels(token["children"])


class MkdocstringsExtension(Extension):
    """Our Markdown extension.

    It cannot work outside of `mkdocstrings`.
    """

    def __init__(
        self,
        handlers: Handlers,
        autorefs: AutorefsPlugin,
        *,
        autorefs_extension: bool = False,
        **kwargs: Any,
    ) -> None:
        """Initialize the object.

        Arguments:
            handlers: The handlers container.
            autorefs: The autorefs plugin instance.
            autorefs_extension: Whether the autorefs extension must be registered.
            **kwargs: Keyword arguments used by `markdown.extensions.Extension`.
        """
        super().__init__(**kwargs)
        self._handlers = handlers
        self._autorefs = autorefs
        self._autorefs_extension = autorefs_extension

    def extendMarkdown(self, md: Markdown) -> None:  # noqa: N802 (casing: parent method's name)
        """Register the extension.

        Add an instance of our [`AutoDocProcessor`][mkdocstrings.AutoDocProcessor] to the Markdown parser.

        Arguments:
            md: A `markdown.Markdown` instance.
        """
        md.registerExtension(self)

        # Zensical integration: get the current page from the Zensical-specific preprocessor.
        if "zensical_current_page" in md.preprocessors:
            self._autorefs.current_page = md.preprocessors["zensical_current_page"]

        md.parser.blockprocessors.register(
            AutoDocProcessor(md, handlers=self._handlers, autorefs=self._autorefs),
            "mkdocstrings",
            priority=75,  # Right before markdown.blockprocessors.HashHeaderProcessor
        )
        md.treeprocessors.register(
            _HeadingsPostProcessor(md),
            "mkdocstrings_post_headings",
            priority=4,  # Right after 'toc'.
        )
        md.treeprocessors.register(
            _TocLabelsTreeProcessor(md),
            "mkdocstrings_post_toc_labels",
            priority=4,  # Right after 'toc'.
        )

        if self._autorefs_extension:
            AutorefsExtension(self._autorefs).extendMarkdown(md)


# -----------------------------------------------------------------------------
# The following is only used by Zensical. The goal is to provide temporary
# compatibility for users migrating from MkDocs (and Material for MkDocs)
# to Zensical. When detecting the use of the mkdocstrings plugin in mkdocs.yml,
# Zensical will add the mkdocstrings extension to its Markdown extensions.

_default_config: dict[str, Any] = {
    "default_handler": "python",
    "handlers": {},
    "custom_templates": None,
    "locale": "en",
    "enable_inventory": True,
    "enabled": True,
}


def _split_configs(
    markdown_extensions: list[str | dict[str, dict[str, Any]] | Extension],
) -> tuple[list[str | Extension], dict[str, dict[str, Any]]]:
    # Split markdown extensions and their configs from mkdocs.yml
    mdx: list[str | Extension] = []
    mdx_config: dict[str, dict[str, Any]] = {}
    for item in markdown_extensions:
        if isinstance(item, (str, Extension)):
            mdx.append(item)
        elif isinstance(item, dict):
            for key, value in item.items():
                mdx.append(key)
                mdx_config[key] = value
                break  # Only one item per dict
    return mdx, mdx_config


class _ToolConfig:
    def __init__(self, config_file_path: str | None = None) -> None:
        self.config_file_path = config_file_path


_AUTOREFS = None
_HANDLERS = None


def makeExtension(  # noqa: N802
    *,
    default_handler: str | None = None,
    inventory_project: str | None = None,
    inventory_version: str | None = None,
    handlers: dict[str, dict] | None = None,
    custom_templates: str | None = None,
    markdown_extensions: list[str | dict | Extension] | None = None,
    locale: str | None = None,
    config_file_path: str | None = None,
) -> MkdocstringsExtension:
    """Create the extension instance.

    We only support this function being used by Zensical.
    Consider this function private API.
    """
    global _AUTOREFS  # noqa: PLW0603
    if _AUTOREFS is None:
        _AUTOREFS = AutorefsPlugin()
        _AUTOREFS.config = AutorefsConfig()  # ty:ignore[invalid-assignment]
        _AUTOREFS.config.resolve_closest = True
        _AUTOREFS.config.link_titles = "auto"
        _AUTOREFS.config.strip_title_tags = "auto"
        _AUTOREFS.scan_toc = True
        _AUTOREFS._link_titles = "external"
        _AUTOREFS._strip_title_tags = False

    global _HANDLERS  # noqa: PLW0603
    if _HANDLERS is None:
        mdx, mdx_config = _split_configs(markdown_extensions or [])
        tool_config = _ToolConfig(config_file_path=config_file_path)
        mdx.append(AutorefsExtension(_AUTOREFS))
        _HANDLERS = Handlers(
            theme="material",
            default=default_handler or _default_config["default_handler"],
            inventory_project=inventory_project or "Project",
            inventory_version=inventory_version or "0.0.0",
            handlers_config=handlers or _default_config["handlers"],
            custom_templates=custom_templates or _default_config["custom_templates"],
            mdx=mdx,
            mdx_config=mdx_config,
            locale=locale or _default_config["locale"],
            tool_config=tool_config,
        )

        _HANDLERS._download_inventories()
        register = _AUTOREFS.register_url
        for identifier, url in _HANDLERS._yield_inventory_items():
            register(identifier, url)

    return MkdocstringsExtension(
        handlers=_HANDLERS,
        autorefs=_AUTOREFS,
        autorefs_extension=True,
    )


def _reset() -> None:
    global _AUTOREFS, _HANDLERS  # noqa: PLW0603
    _AUTOREFS = None
    _HANDLERS = None


def _get_autorefs() -> dict[str, Any]:
    if _AUTOREFS:
        return {
            "primary": _AUTOREFS._primary_url_map,
            "secondary": _AUTOREFS._secondary_url_map,
            "inventory": _AUTOREFS._abs_url_map,
            "titles": _AUTOREFS._title_map,
        }
    return {}


def _get_inventory() -> bytes:
    if _HANDLERS:
        return _HANDLERS.inventory.format_sphinx()
    return b""



"""Tests for the extension module."""

from __future__ import annotations

import re
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from markdown import Markdown

    from mkdocstrings import MkdocstringsPlugin


@pytest.mark.parametrize("ext_markdown", [{"markdown_extensions": [{"footnotes": {}}]}], indirect=["ext_markdown"])
def test_multiple_footnotes(ext_markdown: Markdown) -> None:
    """Assert footnotes don't get added to subsequent docstrings."""
    output = ext_markdown.convert(
        dedent(
            """
            Top.[^aaa]

            ::: tests.fixtures.footnotes.func_a

            ::: tests.fixtures.footnotes.func_b

            ::: tests.fixtures.footnotes.func_c

            [^aaa]: Top footnote
            """,
        ),
    )
    assert output.count("Footnote A") == 1
    assert output.count("Footnote B") == 1
    assert output.count("Top footnote") == 1


def test_markdown_heading_level(ext_markdown: Markdown) -> None:
    """Assert that Markdown headings' level doesn't exceed heading_level."""
    output = ext_markdown.convert("::: tests.fixtures.headings\n    options:\n      show_root_heading: true")
    assert ">Foo</h3>" in output
    assert ">Bar</h5>" in output
    assert ">Baz</h6>" in output


def test_keeps_preceding_text(ext_markdown: Markdown) -> None:
    """Assert that autodoc is recognized in the middle of a block and preceding text is kept."""
    output = ext_markdown.convert("**preceding**\n::: tests.fixtures.headings")
    assert "<strong>preceding</strong>" in output
    assert ">Foo</h2>" in output
    assert ":::" not in output


def test_reference_inside_autodoc(ext_markdown: Markdown) -> None:
    """Assert cross-reference Markdown extension works correctly."""
    output = ext_markdown.convert("::: tests.fixtures.cross_reference")
    assert re.search(r"Link to <.*something\.Else.*>something\.Else<.*>\.", output)


def test_quote_inside_annotation(ext_markdown: Markdown) -> None:
    """Assert that inline highlighting doesn't double-escape HTML."""
    output = ext_markdown.convert("::: tests.fixtures.string_annotation.Foo")
    assert ";hi&" in output
    assert "&amp;" not in output


def test_html_inside_heading(ext_markdown: Markdown) -> None:
    """Assert that headings don't double-escape HTML."""
    output = ext_markdown.convert("::: tests.fixtures.html_tokens")
    assert "&#39;&lt;" in output
    assert "&amp;" not in output


@pytest.mark.parametrize(
    ("ext_markdown", "expect_permalink"),
    [
        ({"markdown_extensions": [{"toc": {"permalink": "@@@"}}]}, "@@@"),
        ({"markdown_extensions": [{"toc": {"permalink": "TeSt"}}]}, "TeSt"),
        ({"markdown_extensions": [{"toc": {"permalink": True}}]}, "&para;"),
    ],
    indirect=["ext_markdown"],
)
def test_no_double_toc(ext_markdown: Markdown, expect_permalink: str) -> None:
    """Assert that the 'toc' extension doesn't apply its modification twice."""
    output = ext_markdown.convert(
        dedent(
            """
            # aa

            ::: tests.fixtures.headings
                options:
                    show_root_toc_entry: false

            # bb
            """,
        ),
    )
    assert output.count(expect_permalink) == 5
    assert 'id="tests.fixtures.headings--foo"' in output
    assert ext_markdown.toc_tokens == [  # ty: ignore[unresolved-attribute]
        {
            "level": 1,
            "id": "aa",
            "html": "aa",
            "name": "aa",
            "data-toc-label": "",
            "children": [
                {
                    "level": 2,
                    "id": "tests.fixtures.headings--foo",
                    "html": "Foo",
                    "name": "Foo",
                    "data-toc-label": "",
                    "children": [
                        {
                            "level": 4,
                            "id": "tests.fixtures.headings--bar",
                            "html": "Bar",
                            "name": "Bar",
                            "data-toc-label": "",
                            "children": [
                                {
                                    "level": 6,
                                    "id": "tests.fixtures.headings--baz",
                                    "html": "Baz",
                                    "name": "Baz",
                                    "data-toc-label": "",
                                    "children": [],
                                },
                            ],
                        },
                    ],
                },
            ],
        },
        {
            "level": 1,
            "id": "bb",
            "html": "bb",
            "name": "bb",
            "data-toc-label": "",
            "children": [],
        },
    ]


def test_use_custom_handler(ext_markdown: Markdown) -> None:
    """Assert that we use the custom handler declared in an individual autodoc instruction."""
    with pytest.raises(ModuleNotFoundError):
        ext_markdown.convert("::: tests.fixtures.headings\n    handler: not_here")


def test_register_every_identifier_alias(plugin: MkdocstringsPlugin, ext_markdown: Markdown) -> None:
    """Assert that we don't preemptively register all identifiers of a rendered object."""
    handler = plugin._handlers.get_handler("python")  # ty: ignore[unresolved-attribute]
    ids = ("id1", "id2", "id3")
    handler.get_aliases = lambda _: ids  # ty: ignore[invalid-assignment]
    autorefs = ext_markdown.parser.blockprocessors["mkdocstrings"]._autorefs

    class Page:
        url = "foo"

    autorefs.current_page = Page()
    ext_markdown.convert("::: tests.fixtures.headings")
    for identifier in ids:
        assert identifier in autorefs._secondary_url_map


def test_use_options_yaml_key(ext_markdown: Markdown) -> None:
    """Check that using the 'options' YAML key works as expected."""
    assert "h1" in ext_markdown.convert("::: tests.fixtures.headings\n    options:\n      heading_level: 1")
    assert "h1" not in ext_markdown.convert("::: tests.fixtures.headings\n    options:\n      heading_level: 2")


def test_use_yaml_options_after_blank_line(ext_markdown: Markdown) -> None:
    """Check that YAML options are detected even after a blank line."""
    assert "h1" not in ext_markdown.convert("::: tests.fixtures.headings\n\n    options:\n      heading_level: 2")


@pytest.mark.parametrize("ext_markdown", [{"markdown_extensions": [{"admonition": {}}]}], indirect=["ext_markdown"])
def test_removing_duplicated_headings(ext_markdown: Markdown) -> None:
    """Assert duplicated headings are removed from the output."""
    output = ext_markdown.convert(
        dedent(
            """
            ::: tests.fixtures.headings_many.heading_1

            !!! note

                ::: tests.fixtures.headings_many.heading_2

            ::: tests.fixtures.headings_many.heading_3
            """,
        ),
    )
    assert output.count(">Heading one<") == 1
    assert output.count(">Heading two<") == 1
    assert output.count(">Heading three<") == 1


def _assert_contains_in_order(items: list[str], string: str) -> None:
    index = 0
    for item in items:
        assert item in string[index:]
        index = string.index(item, index) + len(item)


@pytest.mark.parametrize("ext_markdown", [{"markdown_extensions": [{"attr_list": {}}]}], indirect=["ext_markdown"])
def test_backup_of_anchors(ext_markdown: Markdown) -> None:
    """Anchors with empty `href` are backed up."""
    output = ext_markdown.convert("::: tests.fixtures.markdown_anchors")

    # Anchors with id and no href have been backed up and updated.
    _assert_contains_in_order(
        [
            'id="anchor"',
            'id="tests.fixtures.markdown_anchors--anchor"',
            'id="heading-anchor-1"',
            'id="tests.fixtures.markdown_anchors--heading-anchor-1"',
            'id="heading-anchor-2"',
            'id="tests.fixtures.markdown_anchors--heading-anchor-2"',
            'id="heading-anchor-3"',
            'id="tests.fixtures.markdown_anchors--heading-anchor-3"',
        ],
        output,
    )

    # Anchors with href and with or without id have been updated but not backed up.
    _assert_contains_in_order(
        [
            'id="tests.fixtures.markdown_anchors--with-id"',
        ],
        output,
    )
    assert 'id="with-id"' not in output

    _assert_contains_in_order(
        [
            'href="#tests.fixtures.markdown_anchors--has-href1"',
            'href="#tests.fixtures.markdown_anchors--has-href2"',
        ],
        output,
    )
    assert 'href="#has-href1"' not in output
    assert 'href="#has-href2"' not in output



"""Tests for the inventory module."""

from __future__ import annotations

from io import BytesIO
from os.path import join
from pathlib import Path

import pytest
from mkdocs.commands.build import build
from mkdocs.config import load_config

from mkdocstrings import Inventory, InventoryItem


@pytest.mark.parametrize(
    "our_inv",
    [
        Inventory(),
        Inventory([InventoryItem(name="object_path", domain="py", role="obj", uri="page_url")]),
        Inventory([InventoryItem(name="object_path", domain="py", role="obj", uri="page_url#object_path")]),
        Inventory([InventoryItem(name="object_path", domain="py", role="obj", uri="page_url#other_anchor")]),
        Inventory([InventoryItem(name="o", domain="py", role="obj", uri="u#o", dispname="first line\nsecond line")]),
    ],
)
def test_sphinx_load_inventory_file(our_inv: Inventory) -> None:
    """Perform the 'live' inventory load test."""
    sphinx = pytest.importorskip("sphinx.util.inventory", reason="Sphinx is not installed")

    buffer = BytesIO(our_inv.format_sphinx())
    sphinx_inv = sphinx.InventoryFile.load(buffer, "", join)

    sphinx_inv_length = sum(len(sphinx_inv[key]) for key in sphinx_inv)
    assert sphinx_inv_length == len(our_inv.values())

    for item in our_inv.values():
        assert item.name in sphinx_inv[f"{item.domain}:{item.role}"]


def test_sphinx_load_mkdocstrings_inventory_file() -> None:
    """Perform the 'live' inventory load test on mkdocstrings own inventory."""
    sphinx = pytest.importorskip("sphinx.util.inventory", reason="Sphinx is not installed")

    mkdocs_config = load_config()
    mkdocs_config["plugins"].run_event("startup", command="build", dirty=False)
    try:
        build(mkdocs_config)
    finally:
        mkdocs_config["plugins"].run_event("shutdown")
    own_inv = mkdocs_config["plugins"]["mkdocstrings"].handlers.inventory

    with Path("site/objects.inv").open("rb") as fp:
        sphinx_inv = sphinx.InventoryFile.load(fp, "", join)

    sphinx_inv_length = sum(len(sphinx_inv[key]) for key in sphinx_inv)
    assert sphinx_inv_length == len(own_inv.values())

    for item in own_inv.values():
        assert item.name in sphinx_inv[f"{item.domain}:{item.role}"]


@pytest.mark.parametrize(
    "our_inv",
    [
        Inventory(),
        Inventory([InventoryItem(name="object_path", domain="py", role="obj", uri="page_url")]),
        Inventory([InventoryItem(name="object_path", domain="py", role="obj", uri="page_url#object_path")]),
        Inventory([InventoryItem(name="object_path", domain="py", role="obj", uri="page_url#other_anchor")]),
        Inventory([InventoryItem(name="o", domain="py", role="obj", uri="u#o", dispname="first line\nsecond line")]),
    ],
)
def test_mkdocstrings_roundtrip_inventory_file(our_inv: Inventory) -> None:
    """Save some inventory files, then load them in again."""
    buffer = BytesIO(our_inv.format_sphinx())
    round_tripped = Inventory.parse_sphinx(buffer)

    assert our_inv.keys() == round_tripped.keys()
    for key, value in our_inv.items():
        round_tripped_item = round_tripped[key]
        assert round_tripped_item.name == value.name
        assert round_tripped_item.domain == value.domain
        assert round_tripped_item.role == value.role
        assert round_tripped_item.uri == value.uri
        assert round_tripped_item.priority == value.priority
        assert round_tripped_item.dispname == value.dispname.splitlines()[0]


def test_parse_inventory_project_and_version() -> None:
    """Parse project and version in inventory header."""
    base_inv = Inventory(project="hello", version="1000")
    parsed_inv = Inventory.parse_sphinx(BytesIO(base_inv.format_sphinx()))
    assert base_inv.project == parsed_inv.project
    assert base_inv.version == parsed_inv.version
